/*----------------------------------------------------------------------------+
 |                                                                            |
 |                       HEYU Configuration                                   |
 |              Copyright 2002,2003,2004 Charles W. Sullivan                  |
 |                      All Right Reserved                                    |
 |                                                                            |
 |                                                                            |
 | This software is licensed free of charge for non-commercial distribution   |
 | and for personal and internal business use only.  Inclusion of this        |
 | software or any part thereof in a commercial product is prohibited         |
 | without the prior written permission of the author.  You may copy, use,    |
 | and distribute this software subject to the following restrictions:        |
 |                                                                            |
 |  1)	You may not charge money for it.                                      |
 |  2)	You may not remove or alter this license, copyright notice, or the    |
 |      included disclaimers.                                                 |
 |  3)	You may not claim you wrote it.                                       |
 |  4)	If you make improvements (or other changes), you are requested        |
 |	to send them to the official Heyu maintainer, Daniel B. Suthers       |
 |      (dbs@tanj.com), so there's a focal point for distributing improved    |
 |      versions.                                                             |
 |                                                                            |
 | As used herein, HEYU is a trademark of Daniel B. Suthers, while X10,       | 
 | CM11A, and Activehome are trademarks of X10 Wireless Technology, Inc.      |
 | The author is not affiliated with either entity.                           |
 |                                                                            |
 | Charles W. Sullivan                                                        |
 | Greensboro, North Carolina                                                 |
 | Email: cwsulliv@triad.rr.com                                               |
 |                                                                            |
 | Disclaimers:                                                               |
 | THERE IS NO ASSURANCE THAT THIS SOFTWARE IS FREE OF DEFECTS AND IT MUST    |
 | NOT BE USED IN ANY SITUATION WHERE THERE IS ANY CHANCE THAT ITS            |
 | PERFORMANCE OR FAILURE TO PERFORM AS EXPECTED COULD RESULT IN LOSS OF      |
 | LIFE, INJURY TO PERSONS OR PROPERTY, FINANCIAL LOSS, OR LEGAL LIABILITY.   |
 |                                                                            |
 | TO THE EXTENT ALLOWED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED "AS IS",|
 | WITH NO EXPRESS OR IMPLIED WARRANTY, INCLUDING, BUT NOT LIMITED TO, THE    |
 | IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.|
 |                                                                            |
 | IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL THE AUTHOR BE LIABLE    |
 | FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL   |
 | DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SOFTWARE EVEN IF   |
 | THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.            |
 |                                                                            |
 +----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#if defined(SYSV) || defined(FREEBSD)
#include <string.h>
#else
#include <strings.h>
#endif
#include "x10.h"
#include "process.h"

extern int line_no;
extern char heyu_path[PATH_LEN + 1];
extern char alt_path[PATH_LEN + 1];
extern char heyu_script[PATH_LEN + 1];
extern char statefile[PATH_LEN + 1];
extern long std_tzone;
extern char default_housecode;

CONFIG config;

enum {
   OldAlias, Alias, UserSyn, Launcher, Script, Scene,
   /* Directives with multiple instances allowed */
   /* must be before 'Separator'.                */ 
   Separator,  
   Tty, HouseCode, ForceAddr,
   NewFormat, ScheduleFile, MaxPParms, RcsTemp, StatusTimeout,
   SpfTimeout, DispExpMac, AckHails, Mode, ProgramDays,
   AsIfDate, AsIfTime, CombineEvents, CompressMacros, FebKluge,
   Latitude, Longitude, DawnOption, DuskOption, DawnSubstitute,
   DuskSubstitute, MinDawn, MaxDawn, MinDusk, MaxDusk, CheckFiles,
   ReportPath, ReplDelay, Ignored, ResvTimers, TrigTag, XrefApp,
   MacTerm, AutoChain, ResOverlap, ModuleTypes, LaunchMode,
   FunctionMode, LaunchSrc, DefaultModule, ScriptShell, ScriptMode,
   LogDir,IsDarkOffset,EnvAliasPrefix,
};

static struct conf {
   char          *name;
   int           mintoken;
   int           maxtoken;
   unsigned char isparsed;
   unsigned char override;  /* Allow override in sched file */
   unsigned char switchval;
} command[] = {
   {"(_alias_)",            3, 3, 0, 0, OldAlias },
   {"ALIAS",                3, 6, 0, 0, Alias },
   {"LAUNCHER",             2,-1, 0, 0, Launcher},
   {"SCRIPT",               2,-1, 0, 0, Script },
   {"USERSYN",              2,-1, 0, 0, UserSyn },
   {"SCENE",                2,-1, 0, 0, Scene },
   /* Multiple instances OK for switchval above */
   {"TTY",                  2, 2, 0, 0, Tty },
   {"HOUSECODE",            2, 2, 0, 0, HouseCode },
   {"FORCE_ADDR",           2, 2, 0, 0, ForceAddr },
   {"NEWFORMAT",            1, 1, 0, 0, Ignored },
   {"MACROXREF",            2, 2, 0, 0, Ignored },
   {"SCHEDULE_FILE",        2, 2, 0, 0, ScheduleFile },
   {"MAX_PPARMS",           2, 2, 0, 0, MaxPParms },
   {"STATUS_TIMEOUT",       2, 2, 0, 0, StatusTimeout },
   {"SPF_TIMEOUT",          2, 2, 0, 0, SpfTimeout },
   {"RCS_TEMPERATURE",      2, 2, 0, 0, RcsTemp },
   {"RCS_DECODE",           2, 2, 0, 0, RcsTemp },
   {"DISPLAY_EXP_MACROS",   2, 2, 0, 0, Ignored }, 
   {"ACK_HAILS",            2, 2, 0, 0, AckHails }, 
   {"MODE",                 2, 2, 0, 1, Mode },
   {"PROGRAM_DAYS",         2, 2, 0, 1, ProgramDays },
   {"ASIF_DATE",            2, 2, 0, 1, AsIfDate },
   {"ASIF_TIME",            2, 2, 0, 1, AsIfTime },
   {"COMBINE_EVENTS",       2, 2, 0, 1, CombineEvents },
   {"COMPRESS_MACROS",      2, 2, 0, 1, CompressMacros },
   {"FEB_KLUGE",            2, 2, 0, 1, FebKluge },
   {"LATITUDE",             2, 2, 0, 1, Latitude },
   {"LONGITUDE",            2, 2, 0, 1, Longitude },
   {"DAWN_OPTION",          2, 2, 0, 1, DawnOption },
   {"DUSK_OPTION",          2, 2, 0, 1, DuskOption },
   {"DAWN_SUBSTITUTE",      2, 2, 0, 1, Ignored },
   {"DUSK_SUBSTITUTE",      2, 2, 0, 1, Ignored },
   {"MIN_DAWN",             2, 2, 0, 1, MinDawn },
   {"MAX_DAWN",             2, 2, 0, 1, MaxDawn },
   {"MIN_DUSK",             2, 2, 0, 1, MinDusk },
   {"MAX_DUSK",             2, 2, 0, 1, MaxDusk },
   {"WRITE_CHECK_FILES",    2, 2, 0, 1, CheckFiles },
   {"REPORT_PATH",          2, 2, 0, 1, ReportPath },
   {"REPL_DELAYED_MACROS",  2, 2, 0, 1, ReplDelay },
   {"RESERVED_TIMERS",      2, 2, 0, 1, ResvTimers }, /* WIP */
   {"TRIGGER_TAG",          2, 2, 0, 1, Ignored }, 
   {"XREF_APPEND",          2, 2, 0, 1, Ignored }, 
   {"MACTERM",              2, 2, 0, 1, MacTerm }, /* WIP */
   {"RESOLVE_OVERLAP",      2, 2, 0, 1, ResOverlap },
   {"SCRIPT_MODE",          2, 2, 0, 0, ScriptMode },
   {"SCRIPT_SHELL",         2, 2, 0, 0, ScriptShell },
   {"LAUNCH_SOURCE",        2, 6, 0, 0, LaunchSrc },
   {"DEFAULT_MODULE",       2, 2, 0, 0, DefaultModule },
   {"LOG_DIR",              2, 2, 0, 0, LogDir },
   {"ISDARK_OFFSET",        2, 2, 0, 0, IsDarkOffset },
   {"ENV_ALIAS_PREFIX",     2, 2, 0, 0, EnvAliasPrefix },
};
int ncommands = ( sizeof(command)/sizeof(struct conf) );

/* Dawn/Dusk options */
static struct ddopt {
   char     *label;
   unsigned char value;
} dd_option[] = {
   {"FIRST",    FIRST },
   {"EARLIEST", EARLIEST },
   {"LATEST",   LATEST },
   {"AVERAGE",  AVERAGE },
   {"MEDIAN",   MEDIAN },
};
int nddopt = ( sizeof(dd_option)/sizeof(struct ddopt) );

/* Launch source options - set default sources of signal */
/* permitted to launch a script.                         */
static struct lsopt {
   char      *label;
   unsigned char value;
} ls_option[] = {
   {"SNDC",     SNDC   }, /* Sent from command line */
   {"SNDM",     SNDM   }, /* Sent from macro executed by Timer */
   {"SNDT",     SNDT   }, /* Sent from macro executed by Trigger */
   {"RCVI",     RCVI   }, /* Received from the interface */
   {"RCVT",     RCVT   }, /* Trigger which executed a macro */
   {"ANYSRC",   LSALL  }, /* Any of the above (SNDS not included!) */
   {"NOSRC",    LSNONE }, /* No sources are the default */
};
int nlsopt = ( sizeof(ls_option)/sizeof(struct lsopt) );

/*---------------------------------------------------------------------+
 | Reset the isparsed flags in the conf struct array.                  |
 +---------------------------------------------------------------------*/
void reset_isparsed_flags ( void )
{
   int j;

   for ( j = 0; j < ncommands; j++ )
      command[j].isparsed = 0;

   return;
}

/*------------------------------------------------------+
 | Create a suffix for files to be locked by stripping  |
 | the leading "/dev" from the TTY device name and      |
 | replacing all other instances of '/' characters with |
 | the '.' character.                                   |
 +------------------------------------------------------*/
void create_lock_suffix ( char *suffix, char *tty )
{
   char *sp;

   strcpy(suffix, ".");
   if ( strncmp(tty, "/dev/", 5) == 0 )
      strcat(suffix, tty + 5);
   else
      strcat(suffix, tty);

   while ( (sp = strchr(suffix, '/')) )
      *sp = '.';

   return;
}

/*---------------------------------------------------------------------+
 | Parse the users configuration file and save info in CONFIG struct.  |
 | First check to see if it's already been read.                       |
 +---------------------------------------------------------------------*/
int parse_config ( FILE *fd_conf )
{
   char   buffer[LINE_LEN];
   char   *sp1;
   int    err, errors;
   int    j;
   SCENE  *scenep;
   extern char *typename[];

   /* If the configuration file has already been read into memory, */
   /* just return.                                                 */
   if ( config.read_flag != 0 ) {
      return 0;
   }

   /* Make sure the isparsed flags are reset in the command list */
   reset_isparsed_flags();

   /* Load some default configuration values */
   config.read_flag = 0;
   config.mode = DEF_MODE;
   (void) strcpy(config.schedfile, DEF_SCHEDFILE);
   config.asif_date = -1;
   config.asif_time = -1;
   config.program_days = DEF_PROGRAM_DAYS;
   config.combine_events = DEF_COMBINE_EVENTS;
   config.compress_macros = DEF_COMPRESS_MACROS;
   config.feb_kluge = DEF_FEB_KLUGE;
   config.housecode = DEF_HOUSECODE;
   config.force_addr = DEF_FORCE_ADDR;
   config.loc_flag = 0;
   config.dawn_option = DEF_DAWN_OPTION;
   config.dusk_option = DEF_DUSK_OPTION;
   config.dawn_substitute = DEF_DAWN_SUBSTITUTE;
   config.dusk_substitute = DEF_DUSK_SUBSTITUTE;
   config.min_dawn = DEF_MIN_DAWN;
   config.max_dawn = DEF_MAX_DAWN;
   config.min_dusk = DEF_MIN_DUSK;
   config.max_dusk = DEF_MAX_DUSK;
   strcpy(config.tty, DEF_TTY);
   config.newformat = 0;
   config.checkfiles = DEF_CHECK_FILES;
   config.repl_delay = DEF_REPL_DELAYED_MACROS;
   config.reserved_timers = DEF_RESERVED_TIMERS;
   config.aliasp = NULL;
   config.scenep = NULL;
   config.scriptp = NULL;
   config.launcherp = NULL;
   config.max_pparms = DEF_MAX_PPARMS;
   config.rcs_temperature = DEF_RCS_TEMPERATURE;
   config.trigger_tag = DEF_TRIGGER_TAG;
   config.display_offset = DEF_DISPLAY_OFFSET;
   config.xref_append = DEF_XREF_APPEND;
   config.ack_hails = DEF_ACK_HAILS;
   config.macterm = DEF_MACTERM;
   config.status_timeout = DEF_STATUS_TIMEOUT;
   config.spf_timeout = DEF_SPF_TIMEOUT;
   config.disp_exp_mac = DEF_DISPLAY_EXP_MACROS;
   config.module_types = DEF_MODULE_TYPES;
   config.launch_mode = DEF_LAUNCH_MODE;
   config.function_mode = DEF_FUNCTION_MODE;
   config.launch_source = DEF_LAUNCH_SOURCE;
   config.res_overlap = DEF_RES_OVERLAP;
   config.default_module = lookup_module_type(DEF_DEFAULT_MODULE);
   config.script_mode = DEF_SCRIPT_MODE;
   if ( (sp1 = getenv("SHELL")) != NULL )
      strcpy(config.script_shell, sp1);
   else
      strcpy(config.script_shell, "/bin/sh");
   *config.logfile = '\0';
   config.isdark_offset = DEF_ISDARK_OFFSET;
   strcpy(config.env_alias_prefix, DEF_ENV_ALIAS_PREFIX);

   /* Also load the user's timezone */
   get_std_timezone();
   config.tzone = std_tzone;

   /* Determine if Daylight Time is ever in effect and if so,    */
   /* the minutes after midnight it goes into and out of effect. */

   config.isdst = get_dst_info();

   line_no = 0;
   errors = 0;
   while ( fgets(buffer, LINE_LEN, fd_conf) != NULL ) {
      line_no++ ;
      buffer[LINE_LEN - 1] = '\0';

      /* Get rid of comments and blank lines */
      if ( (sp1 = strchr(buffer, '#')) != NULL )
         *sp1 = '\0';
      (void) strtrim(buffer);
      if ( buffer[0] == '\0' )
         continue;

      err = parse_config_tail(buffer, SRC_CONFIG);
      if ( err || *error_message() != '\0' ) { 
         fprintf(stderr, "Config Line %02d: %s\n", line_no, error_message());
         clear_error_message();
      }

      errors += err;
      if ( errors > MAX_ERRORS )
         return 1;
   }

   /* Everything has now been read from the config file and stored */

   /* Verify the syntax of user-defined scenes/usersyns */
   scenep = config.scenep;
   j = 0;
   while ( scenep && scenep[j].line_no > 0 ) {
      if ( verify_scene(scenep[j].body) != 0 ) {
         fprintf(stderr, "Config Line %02d: %s '%s': %s\n",
            scenep[j].line_no, typename[scenep[j].type], scenep[j].label,
            error_message());
         clear_error_message();
         if ( ++errors > MAX_ERRORS )
            return errors;
      }
      else if ( *error_message() != '\0' ) {
         /* Check warning messages */
         fprintf(stderr, "Config Line %02d: %s '%s': %s\n",
            scenep[j].line_no, typename[scenep[j].type], scenep[j].label,
            error_message());
         clear_error_message();
      }
      j++;
   }

   /* Configure module masks */
   set_module_masks( config.aliasp );

   /* Use the Heyu path if the user hasn't provided an alternate */
   if ( !(*alt_path) )
      (void)strcpy(alt_path, heyu_path);

   /* Add configuration items from environment */
   errors += environment_config();

   errors += finalize_config();
   if ( *error_message() != '\0' ) {
      fprintf(stderr, "%s\n", error_message());
      clear_error_message();
   }

   config.read_flag = (errors == 0) ? 1 : 0;

   /* Done with config file */
   line_no = 0; 
  
   return errors;
}
         

/*---------------------------------------------------------------------+
 | Parse the config line in buffer.  Argument 'source' can be          |
 | SRC_CONFIG if called from parse_config() or SRC_SCHED if called     |
 | from parse_sched(), for configuration items which may be overridden |
 | in the latter.                                                      |
 +---------------------------------------------------------------------*/
int parse_config_tail ( char *buffer, unsigned char source ) 
{  
   char   errbuffer[80];
   char   searchstr[128];
   char   directive[128];
   char   label[(NAME_LEN + SCENE_LEN + MACRO_LEN + 1)];
   char   token[50];
   char   hc;
   char   *bufp, *sp;
   int    errors = 0;
   int    num, value, hour, minut, len, modtype;
   int    j, k, commj, switchval;
   int    tokc;
   char   **tokv = NULL;

      bufp = buffer;
      get_token(directive, &bufp, " \t", sizeof(directive)/sizeof(char));
      strtrim(bufp);
      strcpy(searchstr, directive);
      strupper(searchstr);


      /* Search list of configuration commands starting */
      /* past "oldalias".                               */

      commj = 0; switchval = OldAlias;
      for ( j = 1; j < ncommands; j++ ) {
         if ( !strcmp(searchstr, command[j].name) ) {
            switchval = command[j].switchval;
            commj = j;
            break;
         }
      }

      /* See if override in schedule file is permitted */
      if ( (source != SRC_CONFIG) && command[commj].override == 0 ) {
         if ( commj )
            sprintf(errbuffer, "Configuration directive %s not allowed here.",
                command[commj].name);
         else
            sprintf(errbuffer, "Invalid configuration directive");
         store_error_message(errbuffer);
         return 1;
      }

      /* If the directive has a fixed number of tokens (as denoted */
      /* by maxtokens > 0) or _might_ be an old-style alias,       */
      /* tokenize the tail immediately.                            */

      if ( commj == 0 || command[j].maxtoken > 0 ) {
         tokenize(bufp, " \t", &tokc, &tokv);
      }
      else {
         tokc = (*bufp == '\0') ? 0 : 1 ;
      }


      /* If not found in list and it doesn't have the correct */
      /* number of items for an alias, reject it.             */
      if ( commj == 0  && tokc != 2 ) {
         sprintf(errbuffer, "Invalid Directive '%s'", directive);
         store_error_message(errbuffer);
         return 1;
      }

      /* Unless allowed, check to see it's not a duplicate  */
      /* of an earlier directive.                           */
      if ( commj > Separator ) {
         if ( command[j].isparsed & source ) {
            sprintf(errbuffer, 
               "Directive '%s' appears more than once in the file", searchstr);
            store_error_message(errbuffer);
            return 1;
         }
         else
            command[j].isparsed |= source;
      }

      /* Check that commands found on the list have the correct */
      /* number of tokens on the line.                          */
      if ( (tokc + 1) < command[commj].mintoken ) {
         store_error_message("Too few items on line.");
         return 1;
      }
      else if ( (command[commj].maxtoken != -1) && ((tokc + 1) > command[commj].maxtoken) ) {
         store_error_message("Too many items on line.");
         return 1;
      }

      errors = 0;
      switch ( switchval ) {
         case Ignored :
            sprintf(errbuffer,
               "Directive %s is obsolete and is being ignored.", searchstr);
            store_error_message(errbuffer);
            break;

         case Mode :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "COMPATIBLE") )
               config.mode = COMPATIBLE;
            else if ( !strcmp(tokv[0], "HEYU") )
               config.mode = HEYU_MODE;
            else {
               store_error_message("MODE must be COMPATIBLE or HEYU");
               errors++;
            }
            break;

         case OldAlias :  /* Possible alias, else invalid command */
            /* Check if it matches the form of an alias */
            if ( tokc != 2 || strlen(tokv[0]) != 1 ||
                  (hc = toupper(*tokv[0])) < 'A' || hc > 'P' ) {
               store_error_message("Invalid Directive.");
               errors++;
               break;
            }
            if ( strchr(directive, '$') ) {
               store_error_message("Alias labels may not contain the '$' character.");
               errors++;
               break;
            }
            if ( strchr("+-_", *directive ) ) {
               store_error_message("Alias labels may not begin with '+', '-', '_'");
               errors++;
               break;
            }

            if ( add_alias(&config.aliasp, directive, line_no,
                                                 hc, tokv[1], -1) < 0 ) {
               errors++;
            }
            break;

         case Alias :  /* New alias format using ALIAS directive */
            /* Expects housecode and unit code(s) to be concatenated, */
            /* the same as they are for the command line or macro.    */

            if ( strchr(tokv[0], '$') ) {
               store_error_message("Alias labels may not contain the '$' character");
               errors++;
               break;
            }
            if ( strchr("+-_", *tokv[0] ) ) {
               store_error_message("Alias labels may not begin with '+', '-', '_'");
               errors++;
               break;
            }

            hc = toupper( *tokv[1] );
            *tokv[1] = ' ';
            strtrim(tokv[1]);

            modtype = -1;
            if ( tokc > 2 ) {
               if ( (modtype = lookup_module_type(tokv[2])) < 0 ) {
                  sprintf(errbuffer, "Invalid module model '%s'", tokv[2]);
                  store_error_message(errbuffer);
                  errors++;
                  break;
               }
            }

            if ( (j = add_alias(&config.aliasp, tokv[0], line_no,
                                hc, tokv[1], modtype)) < 0 ) {
               errors++;
               break;
            }

            if ( modtype >= 0 && tokc > 3 ) {
               if ( add_module_options(j, tokv + 3, tokc - 3) != 0 ) {
                  errors++;
                  break;
               }
            }
            
            break;

         case UserSyn :
            get_token(label, &bufp, " \t", sizeof(label)/sizeof(char));
            if ( add_scene(&config.scenep, label, line_no, bufp, F_USYN) < 0 ) {
               errors++;
            }
            break;

         case Scene :
            get_token(label, &bufp, " \t", sizeof(label)/sizeof(char));
            if ( add_scene(&config.scenep, label, line_no, bufp, F_SCENE) < 0 ) {
               errors++;
            }
            break;

         case AsIfDate :
            config.asif_date = strtol(tokv[0], NULL, 10);
            if ( config.asif_date < 19700101 ||
                 config.asif_date > 20380101 )  {
               store_error_message("ASIF_DATE must be yyyymmdd between 19700101 and 20380101");
               errors++;
            }
            break;

         case AsIfTime :
            num = sscanf(tokv[0], "%d:%d", &hour, &minut) ;
            value = 60 * hour + minut ;
            config.asif_time = value;
            if ( num != 2 || value < 0 || value > 1439 ) {
               store_error_message("ASIF_TIME must be hh:mm between 00:00-23:59");
               errors++;
            }
            break;

         case ScheduleFile :
            strcpy(config.schedfile, tokv[0]);
            break;

         case ProgramDays :
            strupper(tokv[0]);
            config.program_days = (int)strtol(tokv[0], NULL, 10);

            if ( config.program_days < 1 || config.program_days > 366 ) {
               store_error_message("PROGRAM_DAYS outside range 1-366");
               errors++;
            }
            break;

         case CombineEvents :
            strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.combine_events = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.combine_events = NO;
            else {
               store_error_message("COMBINE_EVENTS must be YES or NO");
               errors++;
            }
            break;

         case CompressMacros :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.compress_macros = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.compress_macros = NO;
            else {
               store_error_message("COMPRESS_MACROS must be YES or NO");
               errors++;
            }
            break;

         case FebKluge :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.feb_kluge = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.feb_kluge = NO;
            else {
               store_error_message("FEB_KLUGE must be YES or NO");
               errors++;
            }
            break;

         case Latitude :            
            errors += parse_latitude(tokv[0]);
            break;

         case Longitude :
            errors += parse_longitude(tokv[0]);
            break;

         case DawnOption :
            (void) strupper(tokv[0]);
            for ( j = 0; j < nddopt; j++ ) {
               if ( !strcmp(tokv[0], dd_option[j].label) )
                  break;
            }
            if ( j == nddopt ) {
               store_error_message("Invalid DAWN_OPTION");
               errors++;
               break;
            }
            config.dawn_option = dd_option[j].value;
            break;

         case DuskOption :
            (void) strupper(tokv[0]);
            for ( j = 0; j < nddopt; j++ ) {
               if ( !strcmp(tokv[0], dd_option[j].label) )
                  break;
            }
            if ( j == nddopt ) {
               store_error_message("Invalid DUSK_OPTION");
               errors++;
               break;
            }
            config.dusk_option = dd_option[j].value;
            break;

         case DawnSubstitute :
            num = sscanf(tokv[0], "%d:%d", &hour, &minut) ;
            value = 60 * hour + minut ;
            if ( num != 2 || value < 0 || value > 1439 ) {
               store_error_message("DAWN_SUBSTITUTE - must be 00:00-23:59 (hh:mm)");
               errors++;
               break;
            }
            config.dawn_substitute = value;
            break;

         case DuskSubstitute :
            num = sscanf(tokv[0], "%d:%d", &hour, &minut) ;
            value = 60 * hour + minut ;
            if ( num != 2 || value < 0 || value > 1439 ) {
               store_error_message("DUSK_SUBSTITUTE must be 00:00-23:59 (hh:mm)");
               errors++;
               break;
            }
            config.dusk_substitute = value;
            break;

         case MinDawn :
            if ( strcmp(strupper(tokv[0]), "OFF") == 0 ) {
               config.min_dawn = OFF;
               break;
            }
            num = sscanf(tokv[0], "%d:%d", &hour, &minut);
            value = 60 * hour + minut ;
            if ( num != 2 || value < 0 || value > 1439 ) {
               store_error_message("MIN_DAWN must be 00:00-23:59 (hh:mm) or OFF");
               errors++;
               break;
            }
            config.min_dawn = value;
            break;

         case MaxDawn :
            if ( strcmp( strupper(tokv[0]), "OFF") == 0 ) {
               config.max_dawn = OFF;
               break;
            }
            num = sscanf(tokv[0], "%d:%d", &hour, &minut);
            value = 60 * hour + minut ;
            if ( num != 2 || value < 0 || value > 1439 ) {
               store_error_message("MAX_DAWN must be 00:00-23:59 (hh:mm) or OFF");
               errors++;
               break;
            }
            config.max_dawn = value;
            break;

         case MinDusk :
            if ( strcmp( strupper(tokv[0]), "OFF") == 0 ) {
               config.min_dusk = OFF;
               break;
            }
            num = sscanf(tokv[0], "%d:%d", &hour, &minut);
            value = 60 * hour + minut ;
            if ( num != 2 || value < 0 || value > 1439 ) {
               store_error_message("MIN_DUSK must be 00:00-23:59 (hh:mm) or OFF");
               errors++;
               break;
            }
            config.min_dusk = value;
            break;

         case MaxDusk :
            if ( strcmp( strupper(tokv[0]), "OFF") == 0 ) {
               config.max_dusk = OFF;
               break;
            }
            num = sscanf(tokv[0], "%d:%d", &hour, &minut);
            value = 60 * hour + minut ;
            if ( num != 2 || value < 0 || value > 1439 ) {
               store_error_message("MAX_DUSK must be 00:00-23:59 (hh:mm) or OFF");
               errors++;
               break;
            }
            config.max_dusk = value;
            break;

         case Tty :
            (void) strcpy(config.tty, tokv[0]);
            break;

         case HouseCode : /* Base housecode */
            hc = toupper(*tokv[0]);
            if ( strlen(tokv[0]) > 1 || hc < 'A' || hc > 'Z' ) {
               store_error_message("Invalid HOUSECODE - must be A though P");
               errors++;
               break;
            }
            config.housecode = hc;
            default_housecode = hc;
            break;

         case ForceAddr :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.force_addr = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.force_addr = NO;
            else {
               store_error_message("FORCE_ADDR must be YES or NO");
               errors++;
            }
            break;

         case NewFormat : /* New Format - allow for additional choices in future */
            if ( (tokc + 1) == 1 )
               config.newformat = 1;
            else
               config.newformat = (unsigned char)strtol(tokv[0], NULL, 10);
            break;

         case CheckFiles :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.checkfiles = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.checkfiles = NO;
            else {
               store_error_message("WRITE_CHECK_FILES must be YES or NO");
               errors++;
            }
            break;

         case ReportPath : /* Alternate path for report files */
            (void)strcpy(alt_path, tokv[0]);
            if ( alt_path[strlen(alt_path) - 1] != '/' )
               (void)strcat(alt_path, "/");
            break;

         case ReplDelay :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.repl_delay = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.repl_delay = NO;
            else {
               store_error_message("REPL_DELAYED_MACROS must be YES or NO");
               errors++;
            }
            break;

         case ResvTimers :
            (void) strupper(tokv[0]);
            config.reserved_timers = (int)strtol(tokv[0], NULL, 10);

            if ( config.reserved_timers < 0 || config.reserved_timers > 50 ) {
               store_error_message("RESERVED_TIMERS outside range 0-50");
               errors++;
            }
            break;

         case TrigTag :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.trigger_tag = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.trigger_tag = NO;
            else {
               store_error_message("TRIGGER_TAG must be YES or NO");
               errors++;
            }
            break;

         case MaxPParms :
            (void) strupper(tokv[0]);
            config.max_pparms = (int)strtol(tokv[0], NULL, 10);

            if ( config.max_pparms < 1 || config.max_pparms > 999 ) {
               store_error_message("MAX_PPARMS outside range 1-999");
               errors++;
            }
            break;

         case RcsTemp :
            /* Housecodes for which Preset commands received from   */
            /* thermostats are to be decoded into temperature.      */
            /* Information is stored as a housecode bitmap.         */
            sp = tokv[0];
            (void) strlower(sp);
            config.rcs_temperature = 0;
            if ( !strcmp(sp, "off") || !strcmp(sp, "no") ) {
               break;
            }
            if ( !strcmp(sp, "all") || !strcmp(sp, "yes") ) {
               config.rcs_temperature = 0xffff;
               break;
            }

            len = strlen(sp);
            if ( *sp == '[' && *(sp + len - 1) == ']' ) {
               /* Create the housecode bitmap */               
               for ( j = 1; j < len - 1; j++ ) {
                  /* Ignore spaces and commas */
                  if ( sp[j] == ' ' || sp[j] == ',' )
                     continue;
                  if ( sp[j] < 'a' || sp[j] > 'p' )
                     break;
                  config.rcs_temperature |= (1 << hc2code(sp[j]));
               }
               if ( j != len - 1 ) {
                  store_error_message("Invalid character in RCS_DECODE housecode list");
                  errors++;
                  break;
               }
               break;
            }
            store_error_message("RCS_DECODE must be OFF or [<housecode list>] or ALL");
            errors++;
            break;

         case StatusTimeout :
            config.status_timeout = (int)strtol(tokv[0], NULL, 10);

            if ( config.status_timeout < 1 || config.status_timeout > 5 ) {
               store_error_message("STATUS_TIMEOUT outside range 1-5");
               errors++;
            }
            break;


         case SpfTimeout :
            config.spf_timeout = (int)strtol(tokv[0], NULL, 10);

            if ( config.spf_timeout < 1 || config.spf_timeout > 10 ) {
               store_error_message("SPF_TIMEOUT outside range 1-10");
               errors++;
            }
            break;

         case XrefApp :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.xref_append = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.xref_append = NO;
            else {
               store_error_message("XREF_APPEND must be YES or NO");
               errors++;
            }
            break;

         case AckHails :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.ack_hails = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.ack_hails = NO;
            else {
               store_error_message("ACK_HAILS must be YES or NO");
               errors++;
            }
            break;

         case DispExpMac :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.disp_exp_mac = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.disp_exp_mac = NO;
            else {
               store_error_message("DISPLAY_EXP_MACROS must be YES or NO");
               errors++;
            }
            break;

         case MacTerm :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.macterm = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.macterm = NO;
            else {
               store_error_message("MACTERM must be YES or NO");
               errors++;
            }
            break;

         case AutoChain :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.auto_chain = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.auto_chain = NO;
            else {
               store_error_message("AUTO_CHAIN must be YES or NO");
               errors++;
            }
            break;

         case Launcher :
            if ( add_launchers(&config.launcherp, line_no, bufp) < 0 )
               errors++;
            break;

         case Script :
            if ( add_script(&config.scriptp, &config.launcherp, line_no, bufp) < 0 )
               errors++;
            break;

         case ModuleTypes :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "YES") )
               config.module_types = YES;
            else if ( !strcmp(tokv[0], "NO") )
               config.module_types = NO;
            else {
               store_error_message("MODULE_TYPES must be YES or NO");
               errors++;
            }
            break;

         case FunctionMode :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "ACTUAL") )
               config.function_mode = FUNC_ACTUAL;
            else if ( !strcmp(tokv[0], "GENERIC") )
               config.function_mode = FUNC_GENERIC;
            else {
               store_error_message("FUNCTION_MODE must be ACTUAL or GENERIC");
               errors++;
            }
            break;

         case LaunchMode :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "SIGNAL") )
               config.launch_mode = TMODE_SIGNAL;
            else if ( !strcmp(tokv[0], "MODULE") )
               config.launch_mode = TMODE_MODULE;
            else {
               store_error_message("LAUNCH_MODE must be SIGNAL or MODULE");
               errors++;
            }
            break;

         case LaunchSrc :
            value = 0;
            for ( j = 0; j < tokc; j++ ) {
               strncpy2(token, tokv[j], sizeof(token));
               strupper(token);
               if ( strcmp(token, "SNDS") == 0 ) {
                  sprintf(errbuffer, "LAUNCH_SOURCE '%s' not allowed as a default.", tokv[j]);
                  store_error_message(errbuffer);
                  errors++;
                  break;
               }
               for ( k = 0; k < nlsopt; k++ ) {
                  if ( strcmp(token, ls_option[k].label) == 0 ) { 
                     value |= ls_option[k].value;
                     break;
                  }
               }
               if ( k >= nlsopt ) {
                  sprintf(errbuffer, "Invalid LAUNCH_SOURCE option '%s'", tokv[j]);
                  store_error_message(errbuffer);
                  errors++;
                  break;
               }
            }
            if ( value & LSNONE ) {
               if ( value & ~LSNONE ) 
                  store_error_message("Warning: LAUNCH_SOURCE 'nosrc' cancels all others on line.");
               config.launch_source = 0;
            }
            else
               config.launch_source = (unsigned char)value;
            break;

         case ResOverlap :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "OLD") )
               config.res_overlap = RES_OVLAP_COMBINED;
            else if ( !strcmp(tokv[0], "NEW") )
               config.res_overlap = RES_OVLAP_SEPARATE;
            else {
               store_error_message("RESOLVE_OVERLAP must be OLD or NEW");
               errors++;
            }
            break;

         case DefaultModule :
            if ( (config.default_module = lookup_module_type(tokv[0])) < 0 ) {
               sprintf(errbuffer, "Module type '%s' is unknown.", tokv[0]);
               store_error_message(errbuffer);
               errors++;
            }
            break;
        
         case ScriptShell :
            if ( access(tokv[0], X_OK) == 0 ) {
              strcpy(config.script_shell, tokv[0]);
            }
            else {
              sprintf(errbuffer,
                 "An executable shell '%s' is not found.", tokv[0]);
              store_error_message(errbuffer);
              errors++;
            }
            break;
         
         case ScriptMode :
            (void) strupper(tokv[0]);
            if ( !strcmp(tokv[0], "HEYUHELPER") )
               config.script_mode = HEYU_HELPER;
            else if ( !strncmp(tokv[0], "SCRIPT", 6) )
               config.script_mode = HEYU_SCRIPT;
            else {
               store_error_message("SCRIPT_MODE must be HEYUHELPER or SCRIPT");
               errors++;
            }
            break;

         case LogDir :
            strncpy2(token, tokv[0], sizeof(token));
            strupper(token);
            if ( strcmp(token, "NONE") == 0 )
               *config.logfile = '\0';
            else {
               sprintf(config.logfile, "%s/%s",
		 tokv[0], LOGFILE);
	    }
            break;

         case IsDarkOffset :
            config.isdark_offset = (int)strtol(tokv[0], NULL, 10);

            if ( config.isdark_offset < 1 || config.isdark_offset > 360 ) {
               store_error_message("ISDARK_OFFSET outside range 1-360 minutes");
               errors++;
            }
            break;

	 case EnvAliasPrefix :
	    strupper(tokv[0]);
	    if ( strcmp(tokv[0], "UC") == 0 )
               strcpy(config.env_alias_prefix, "X10");
	    else if ( strcmp(tokv[0], "LC") == 0 )
	       strcpy(config.env_alias_prefix, "x10");
	    else {
	       store_error_message("ENV_ALIAS_PREFIX must be UC or LC");
	       errors++;
	    }
	    break;

      }
      free( tokv );

      return errors;
}

/*---------------------------------------------------------------------+
 | Get certain configuration items from environment.                   |
 +---------------------------------------------------------------------*/
int environment_config ( void )
{
   int  j, retcode, errors = 0;
   char *sp;
   char buffer[32];

   static char *envars[] = {
     "LATITUDE",
     "LONGITUDE",
     "ASIF_DATE",
     "ASIF_TIME",
   };

   reset_isparsed_flags();

   for ( j = 0; j < (sizeof(envars)/sizeof(char *)); j++ ) {
      if ( (sp = getenv(envars[j])) != NULL ) {
         sprintf(buffer, "%s %s", envars[j], sp);
         retcode = parse_config_tail(buffer, SRC_ENVIRON);
         errors += retcode;
         if ( retcode != 0 || *error_message() != '\0' ) {
            fprintf(stderr, "Environment variable %s.\n", error_message());
            clear_error_message();
         }
      }
   }
   return errors;
}


/*---------------------------------------------------------------------+
 | Display configuration file directives that have been overridden,    |
 | either by an environment variable or by a 'config' command in the   |
 | schedule file, as indicated by the isparsed flag.                   |
 +---------------------------------------------------------------------*/
void display_config_overrides ( FILE *fd )
{
   int  j, count = 0;
   char delim = ' ';

   fprintf(fd, "Configuration overrides:");

   for ( j = 0; j < ncommands; j++ ) {
      if ( command[j].isparsed & (SRC_ENVIRON | SRC_SCHED)) {
         count++;
         fprintf(fd, "%c %s", delim, command[j].name);
         delim = ',';
      }
   }
   if ( !count )
      fprintf(fd, " -- None --");

   fprintf(fd, "\n\n");

   return;
}

/*---------------------------------------------------------------------+
 | Return the index in the array of SCRIPT structures for the SCRIPT   |
 | having the argument 'label', or -1 if not found.                    |
 +---------------------------------------------------------------------*/
int lookup_script ( SCRIPT *scriptp, char *label )
{
   int    j = 0;

   while ( scriptp && scriptp[j].line_no > 0 ) {
      if ( strcmp(label, scriptp[j].label) == 0 )
         return j;
      j++;
   }
   return -1;
}

/*---------------------------------------------------------------------+
 | Return the index in the array of LAUNCHER structures for the first  |
 | LAUNCHER having the argument 'label', or -1 if not found.           |
 +---------------------------------------------------------------------*/
int lookup_launcher ( LAUNCHER *launcherp, char *label )
{
   int   j = 0;

   while ( launcherp && launcherp[j].line_no > 0 ) {
      if ( strcmp(label, launcherp[j].label) == 0 )
         return j;
      j++;
   }
   return -1;
}


/*---------------------------------------------------------------------+
 | Resolve interrelated configuration items.                           |
 +---------------------------------------------------------------------*/
int finalize_config ( void )
{
   int finalize_launchers(void);

   if ( config.mode == COMPATIBLE ) {
      config.program_days = 366;
   }

 
   if ( config.min_dusk != OFF && config.max_dusk != OFF && 
        config.min_dusk >= config.max_dusk ) {
      store_error_message("MIN_DUSK must be less than MAX_DUSK");
      return 1;
   }

   if ( config.min_dawn != OFF && config.max_dawn != OFF && 
        config.min_dawn >= config.max_dawn ) {
      store_error_message("MIN_DAWN must be less than MAX_DAWN");
      return 1;
   }

   create_lock_suffix(config.suffix, config.tty);

   if ( *config.logfile == '\0' ) {
      strcpy(config.logfile, DEF_LOGFILE);
   }
   else {
      strcat(config.logfile, config.suffix);
   }

   if ( finalize_launchers() > 0 )
      return 1;
   
   return 0;
}


/*---------------------------------------------------------------------+
 | Add an alias to the array of ALIAS structures and return the index  |
 | of the new member.                                                  |
 +---------------------------------------------------------------------*/
int add_alias ( ALIAS **aliaspp, char *label, int line_no, 
                              char housecode, char *units, int modtype )
{
   static int    size, max_size;
   static int    strucsize = sizeof(ALIAS);
   int           j, maxlevel;
   int           blksize = 10;
   unsigned char hc;
   unsigned int  bmap;
   unsigned long flags;
   char          errmsg[128];

   clear_error_message();

   /* Allocate initial block of memory */
   if ( *aliaspp == NULL ) {
      *aliaspp = (ALIAS *) calloc(blksize, strucsize );
      if ( *aliaspp == NULL ) {
         (void) fprintf(stderr, "Unable to allocate memory for Alias.\n");
         exit(1);
      }
      max_size = blksize;
      size = 0;
      /* Initialize it where necessary */
      for ( j = 0; j < max_size; j++ ) {
         (*aliaspp)[j].label[0] = '\0';
         (*aliaspp)[j].modtype = -1;
         (*aliaspp)[j].line_no = 0;
         (*aliaspp)[j].flags = 0;
         (*aliaspp)[j].optflags = 0;
      }
   }

   /* Check for a valid label length */
   if ( strlen(label) > NAME_LEN ) {
      sprintf(errmsg, 
         "Alias label '%s' too long - maximum %d characters", label, NAME_LEN);
      store_error_message(errmsg);
      return -1;
   }

   /* See if the alias label is already in the list.     */
   /* If so, it's an error.                              */
   if ( (j = get_alias(*aliaspp, label, &hc, &bmap)) >= 0 ) {
      (void) sprintf(errmsg, "Duplicate alias label '%s'", label);
      store_error_message(errmsg);
      return -1;
   }

   /* Verify that the label is not 'macro' */
   if ( strcmp(label, "macro") == 0 ) {
      store_error_message("An alias may not have the label 'macro'\n");
      return -1;
   }

   /* Check the housecode */
   if ( housecode == '_' ) {
      (void) sprintf(errmsg, 
         "Alias '%s': Default housecode symbol ('_') is invalid in an alias",
              label);
      store_error_message(errmsg);
      return -1;
   }
   if ( toupper(housecode) < 'A' || toupper(housecode) > 'P' ) {
      (void) sprintf(errmsg, "Alias '%s': Housecode '%c' outside range A-P",
               label, toupper(housecode));
      store_error_message(errmsg);
      return -1;
   }

   /* Check the units list */
   if ( parse_units( units, &bmap ) != 0 ) {
      (void) sprintf(errmsg, "Alias '%s': ", label);
      add_error_prefix(errmsg);
      return -1;
   }

   /* Check to see if there's an available location          */
   /* If not, increase the size of the memory allocation.    */
   /* (Always leave room for a final termination indicator.) */
   if ( size == (max_size - 1)) {
      max_size += blksize ;
      *aliaspp = (ALIAS *) realloc(*aliaspp, max_size * strucsize );
      if ( *aliaspp == NULL ) {
         (void) fprintf(stderr, "Unable to increase size of Alias list.\n");
         exit(1);
      }

      /* Initialize the new memory allocation */
      for ( j = size; j < max_size; j++ ) {
         (*aliaspp)[j].label[0] = '\0';
         (*aliaspp)[j].modtype = -1;
         (*aliaspp)[j].line_no = 0;
         (*aliaspp)[j].flags = 0;
         (*aliaspp)[j].optflags = 0;
      }
   }

   j = size;
   size += 1;

   (void) strcpy((*aliaspp)[j].label, label);
   (*aliaspp)[j].line_no = line_no;
   (*aliaspp)[j].housecode = toupper(housecode);
   (*aliaspp)[j].unitbmap = bmap;
   (*aliaspp)[j].modtype = modtype;
   module_attributes(modtype, &flags, &maxlevel);
   (*aliaspp)[j].flags = flags;
   (*aliaspp)[j].maxlevel = maxlevel;
   (*aliaspp)[j].onlevel = maxlevel;

   return j;
}

/*---------------------------------------------------------------------+
 | Add a scene or usersyn to the array of SCENE structures and return  |
 | the index of the new member.                                        |
 | Argument 'type' may be 1 for a scene or 2 for a usersyn             |
 +---------------------------------------------------------------------*/
int add_scene ( SCENE **scenepp, char *label, 
                 int line_no, char *body, unsigned int type )
{
   static int    size, max_size ;
   static int    strucsize = sizeof(SCENE);
   int           j;
   char          *sp;
   int           cmdc, nparms;
   char          **cmdv;
   int           blksize = 10;
   char          errmsg[128];
   extern char   *typename[];

   /* Allocate initial block of memory */
   if ( *scenepp == NULL ) {
      *scenepp = calloc(blksize, strucsize );
      if ( *scenepp == NULL ) {
         fprintf(stderr, "Unable to allocate memory for Scene/Usersyn.\n");
         exit(1);
      }
      max_size = blksize;
      size = 0;
      /* Initialize it */
      for ( j = 0; j < max_size; j++ ) {
         (*scenepp)[j].label[0] = '\0';
         (*scenepp)[j].line_no = -1;
         (*scenepp)[j].nparms = 0;
         (*scenepp)[j].type = 0;
         (*scenepp)[j].body = NULL;
      }
   }

   /* Check for a valid scene label */
   if ( strlen(label) > SCENE_LEN ) {
      sprintf(errmsg, 
         "%s label too long - maximum %d characters", typename[type], SCENE_LEN);
      store_error_message(errmsg);
      return -1;
   }
   if ( strchr("+_-", *label) != NULL ) {
      sprintf(errmsg, "%s label may not may not begin with '+', '-' or '_'", typename[type]);
      store_error_message(errmsg);
      return -1;
   }
   if ( strchr(label, '$') != NULL ) {
      sprintf(errmsg, "%s label may not contain the '$' character", typename[type]);
      store_error_message(errmsg);
      return -1;
   }
   if ( strchr(label, ';') != NULL ) {
      sprintf(errmsg, "%s label may not contain the ';' character", typename[type]);
      store_error_message(errmsg);
      return -1;
   }

   /* See if the scene label is already in the list.     */
   /* If so, it's an error.                              */
   if ( (j = lookup_scene(*scenepp, label)) >= 0 ) {
      sprintf(errmsg, 
         "%s label '%s' previously defined as a %s on line %d",
          typename[type], label, typename[(*scenepp)[j].type], (*scenepp)[j].line_no);
      store_error_message(errmsg);
      return -1;
   }
   if ( is_admin_cmd(label) || is_direct_cmd(label) ) {
      sprintf(errmsg, "%s label '%s' conflicts with heyu command", typename[type], label);
      store_error_message(errmsg);
      return -1;
   }

   /* Check to see if there's an available location          */
   /* If not, increase the size of the memory allocation.    */
   /* (Always leave room for a final termination indicator.) */
   if ( size == (max_size - 1)) {
      max_size += blksize ;
      *scenepp = realloc(*scenepp, max_size * strucsize );
      if ( *scenepp == NULL ) {
         fprintf(stderr, "Unable to increase size of Scene/Usersyn list.\n");
         exit(1);
      }

      /* Initialize the new memory allocation */
      for ( j = size; j < max_size; j++ ) {
         (*scenepp)[j].label[0] = '\0';
         (*scenepp)[j].line_no = -1;
         (*scenepp)[j].nparms = 0;
         (*scenepp)[j].type = 0;
         (*scenepp)[j].body = NULL;
      }
   }

   j = size;
   size += 1;

   /* Determine the number of replaceable parameters */
   sp = strdup(body);
   tokenize(sp, " \t;", &cmdc, &cmdv);
   nparms = max_parms(cmdc, cmdv);
   free(sp);
   free(cmdv);
   if ( *error_message() != '\0' ) {
      sprintf(errmsg, "%s '%s': ", typename[type], label);
      add_error_prefix(errmsg);
   }
   if ( nparms < 0 )
      return -1;

   sp = strdup(body);
   if ( sp == NULL ) {
      fprintf(stderr,
        "Unable to allocate memory for body of Scene/Usersyn '%s'.\n", label);
         exit(1);
   }
   strtrim(sp);
  
   (void) strcpy((*scenepp)[j].label, label);
   (*scenepp)[j].line_no = line_no;
   (*scenepp)[j].nparms = nparms;
   (*scenepp)[j].type = type;
   (*scenepp)[j].body = sp;

   return j;
}

/*---------------------------------------------------------------------+
 | Parse Latitude string [NS+-]ddd:mm and store in struct config.      |
 | line_no <= 0 indicates string from environment, otherwise line in   |
 | configuration file.                                                 | 
 +---------------------------------------------------------------------*/
int parse_latitude ( char *string )
{
   char tmpbuff[128];
   char *sp1, *sp2;
   int  sign;

   if ( string == NULL )
      return 0;

   (void)strncpy2(tmpbuff, string, sizeof(tmpbuff) - 1);
   sp1 = tmpbuff;

   /* For compatibility with Heyu 1 */
   if ( isdigit((int)(*sp1)) ) {
      (void)strcpy(tmpbuff, "N");
      (void)strncpy2(tmpbuff + 1, string, sizeof(tmpbuff) - 2);
   }
   else if ( *sp1 == '+' ) 
      *sp1 = 'N';
   else if ( *sp1 == '-' )
      *sp1 = 'S';
   else ;

   switch ( toupper(*sp1) ) {
      case 'N' :
         sign = 1;
         break;
      case 'S' :
         sign = -1;
         break;
      default :
         sign = -2;   
         break;
   }
   sp1++;

   if ( sign < -1 || !(sp2 = strchr(sp1, ':')) ) {
         store_error_message("LATITUDE invalid - must be [NS+-]dd:mm");
         config.loc_flag &= ~(LATITUDE) ;
         return 1;
   }

   config.lat_m = (int)strtol(sp2 + 1, NULL, 10);
   *sp2 = '\0';
   config.lat_d = (int)strtol(sp1, NULL, 10);
   if ( config.lat_d == 0 )
      config.lat_m *= sign ;
   else
      config.lat_d *= sign ;

   config.latitude = ( config.lat_d < 0 ) ?
         (double)config.lat_d - (double)config.lat_m/60. :
               (double)config.lat_d + (double)config.lat_m/60.;

   if ( config.latitude < -89. || config.latitude > 89. ) {
         store_error_message("LATITUDE outside range -89 to +89 degrees");
         config.loc_flag &= ~(LATITUDE);
         return 1;
   }

   config.loc_flag |= LATITUDE ;

   return 0;
}

/*---------------------------------------------------------------------+
 | Parse Longitude string [EW+-]ddd:mm and store in struct config.     |
 | line_no <= 0 indicates string from environment, otherwise line in   |
 | configuration file.                                                 | 
 +---------------------------------------------------------------------*/
int parse_longitude ( char *string )
{
   char tmpbuff[128];
   char *sp1, *sp2;
   int  sign;

   if ( string == NULL )
      return 0;

   (void)strncpy2(tmpbuff, string, sizeof(tmpbuff) - 1);
   sp1 = tmpbuff;

   /* For compatibility with Heyu 1, where positive longitude assumed West */
   if ( isdigit((int)(*sp1)) ) {
      (void)strcpy(tmpbuff, "W");
      (void)strncpy2(tmpbuff + 1, string, sizeof(tmpbuff) - 2);
   }
   else if ( *sp1 == '+' ) 
      *sp1 = 'W';
   else if ( *sp1 == '-' )
      *sp1 = 'E';
   else ;

   switch ( toupper(*sp1) ) {
      case 'E' :
         sign = 1;
         break;
      case 'W' :
         sign = -1;
         break;
      default :
         sign = -2;     
         break;
   }
   sp1++;

   if ( sign < -1 || !(sp2 = strchr(sp1, ':')) ) {
         store_error_message("LONGITUDE invalid - must be [EW+-]dd:mm");
         config.loc_flag &= ~(LONGITUDE) ;
         return 1;
   }

   config.lon_m = (int)strtol(sp2 + 1, NULL, 10);
   *sp2 = '\0';
   config.lon_d = (int)strtol(sp1, NULL, 10);
   if ( config.lon_d == 0 )
      config.lon_m *= sign ;
   else
      config.lon_d *= sign ;

   config.longitude = ( config.lon_d < 0 ) ?
      (double)config.lon_d - (double)config.lon_m/60. :
               (double)config.lon_d + (double)config.lon_m/60.;

   if ( config.longitude < -180. || config.longitude > 180. ) {
         store_error_message("LONGITUDE outside range -180 to +180 degrees");
         config.loc_flag &= ~(LONGITUDE);
         return 1;
   }

   config.loc_flag |= LONGITUDE ;

   return 0;
}

/*---------------------------------------------------------------------+
 | Check executability of a file on user's PATH                        |
 +---------------------------------------------------------------------*/
int is_executable ( char *pathname )
{
   char pathbuffer[1024];
   char buffer[1024];
   char *sp;
   int  j, tokc;
   char **tokv;

   if ( access(pathname, X_OK) == 0 )
      return 1;

   if ( (sp = getenv("PATH")) == NULL ) 
      return 0;

   strcpy(pathbuffer, sp);

   tokenize(buffer, ":", &tokc, &tokv);

   for ( j = 0; j < tokc; j++ ) {
      strcpy(buffer, tokv[j]);
      strcat(buffer, "/");
      strcat(buffer, pathname);
      if ( access(buffer, X_OK) == 0 ) {
         free(tokv);
         return 1;
      }
   }
   free(tokv);
   return 0;
}

/*---------------------------------------------------------------------+
 | Free memory allocated for SCRIPT structure and contents.            |
 +---------------------------------------------------------------------*/
void free_scripts ( SCRIPT **scriptpp )
{
   int j = 0;

   if ( *scriptpp == NULL )
      return;

   while ( (*scriptpp)[j].line_no > 0 ) {
     if ( (*scriptpp)[j].cmdline ) {
        free((*scriptpp)[j].cmdline);
     }
     j++;
   }
   free(*scriptpp);
   *scriptpp = NULL;

   return;
}  

/*---------------------------------------------------------------------+
 |  Free the array of SCENEs and the scene bodies therein.             |
 +---------------------------------------------------------------------*/
void free_scenes ( SCENE **scenepp )
{
   int j = 0;

   if ( *scenepp == NULL )
      return;

   while ( (*scenepp)[j].line_no > 0 ) {
     if ( (*scenepp)[j].body != NULL ) {
        free((*scenepp)[j].body);
     }
     j++;
   }

   free((*scenepp));
   *scenepp = NULL;
   return;
}

/*---------------------------------------------------------------------+
 |  Free the array of ALIASES.                                         |
 +---------------------------------------------------------------------*/
void free_aliases ( ALIAS **aliaspp )
{
   if ( *aliaspp == NULL )
      return;

   free((*aliaspp));
   *aliaspp = NULL;
   return;
}

/*---------------------------------------------------------------------+
 |  Free the array of LAUNCHERS.                                       |
 +---------------------------------------------------------------------*/
void free_launchers ( LAUNCHER **launcherpp )
{
   if ( *launcherpp == NULL )
      return;

   free((*launcherpp));
   *launcherpp = NULL;
   return;
}

/*---------------------------------------------------------------------+
 |  Free arrays of ALIASes, SCENEs, SCRIPTs, and LAUNCHERs             |
 +---------------------------------------------------------------------*/
void free_all_arrays ( void )
{        
   free_aliases(&config.aliasp);
   free_scenes(&config.scenep);
   free_scripts(&config.scriptp);
   free_launchers(&config.launcherp);

   return;
}

   

   

   

