/*----------------------------------------------------------------------------+
 |                                                                            |
 |              State and Script functions for HEYU                           |
 |              Copyright 2004 Charles W. Sullivan                            |
 |                      All Right Reserved                                    |
 |                                                                            |
 |                                                                            |
 | This software is licensed free of charge for non-commercial distribution   |
 | and for personal and internal business use only.  Inclusion of this        |
 | software or any part thereof in a commercial product is prohibited         |
 | without the prior written permission of the author.  You may copy, use,    |
 | and distribute this software subject to the following restrictions:        |
 |                                                                            |
 |  1)	You may not charge money for it.                                      |
 |  2)	You may not remove or alter this license, copyright notice, or the    |
 |      included disclaimers.                                                 |
 |  3)	You may not claim you wrote it.                                       |
 |  4)	If you make improvements (or other changes), you are requested        |
 |	to send them to the official Heyu maintainer, Daniel B. Suthers       |
 |      (dbs@tanj.com), so there's a focal point for distributing improved    |
 |      versions.                                                             |
 |                                                                            |
 | As used herein, HEYU is a trademark of Daniel B. Suthers, while X10,       | 
 | CM11A, and Activehome are trademarks of X10 Wireless Technology, Inc.      |
 | The author is not affiliated with either entity.                           |
 |                                                                            |
 | Charles W. Sullivan                                                        |
 | Greensboro, North Carolina                                                 |
 | Email: cwsulliv@triad.rr.com                                               |
 |                                                                            |
 | Disclaimers:                                                               |
 | THERE IS NO ASSURANCE THAT THIS SOFTWARE IS FREE OF DEFECTS AND IT MUST    |
 | NOT BE USED IN ANY SITUATION WHERE THERE IS ANY CHANCE THAT ITS            |
 | PERFORMANCE OR FAILURE TO PERFORM AS EXPECTED COULD RESULT IN LOSS OF      |
 | LIFE, INJURY TO PERSONS OR PROPERTY, FINANCIAL LOSS, OR LEGAL LIABILITY.   |
 |                                                                            |
 | TO THE EXTENT ALLOWED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED "AS IS",|
 | WITH NO EXPRESS OR IMPLIED WARRANTY, INCLUDING, BUT NOT LIMITED TO, THE    |
 | IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.|
 |                                                                            |
 | IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL THE AUTHOR BE LIABLE    |
 | FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL   |
 | DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SOFTWARE EVEN IF   |
 | THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.            |
 |                                                                            |
 +----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#if defined(SYSV) || defined(FREEBSD)
#include <string.h>
#else
#include <strings.h>
#endif
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef POSIX
#define EXEC_POSIX
#endif

#ifdef EXEC_POSIX
#include <sys/wait.h>
#endif

#include <time.h>
#include "x10.h"
#include "process.h"


extern CONFIG config;
extern char *funclabel[16];
extern int tty, sptty, verbose;
extern int i_am_monitor, i_am_child, i_am_state;
extern FILE *fdsout, *fdserr;

extern char heyu_script[];
extern char statstr[];
extern char statefile[];
extern char heyu_config[];

#define PROMSIZE 1024
static unsigned char image[PROMSIZE + 10];

extern unsigned int modmask[/* NumModMasks */][16];
extern unsigned char maxdimlevel[16][16];
extern unsigned char ondimlevel[16][16];

/* Generic commands which can launch a script */
static struct trig_type_st {
   char          *label;
   unsigned char generic;
   int           signal;
} trig_type[] = {
  {"on",         0, OnTrig},
  {"off",        0, OffTrig},
  {"dim",        0, DimTrig},
  {"bright",     0, BriTrig},
  {"lightson",   0, LightsOnTrig},
  {"lightsoff",  0, LightsOffTrig},
  {"allon",      0, AllOnTrig},
  {"alloff",     0, AllOffTrig},
  {"status_on",  0, StatusOnTrig},
  {"status_off", 0, StatusOffTrig},
  {"status",     0, StatusReqTrig},
  {"preset",     0, PresetTrig},
  {"extended",   0, ExtendedTrig},
  {"hail",       0, HailReqTrig},
  {"hail_ack",   0, HailAckTrig},
  {"data_xfer",  0, DataXferTrig},
  {"gon",        1, OnTrig},
  {"goff",       1, OffTrig},
  {"gdim",       1, DimTrig},
};
#define NTRIGTYPES (sizeof(trig_type)/sizeof(struct trig_type_st))

/* Functions included in the 'generic' group */
#define GENFUNCMAP ((1 << OnTrig) | (1 << OffTrig) | (1 << DimTrig))

enum {OnState, DimState, ChgState, LightsOnState, NumStates};

static struct x10state_st {
   unsigned char reset;
   unsigned char lastunit;
   unsigned char lastcmd;
   unsigned int  addressed;    /* X10 bitmap */
   unsigned int  state[NumStates]; /* 0 = OnState  1 = DimState  2 = Chgstate 3 = LightsOnState */
   unsigned int  launched;
   unsigned char dimlevel[16]; /* Current dim level */
   unsigned char memlevel[16]; /* "memory" level - preset modules */
   unsigned char lasthc;       /* Only element 0 is used */
} x10state[16];

/*----------------------------------------------------------------------------+
 | Converts programmed dim level (1-22) to dims (0-210).                      |
 | This is just an approximation as the number of dims actually transmitted   |
 | by the interface varies, apparently at random.                             |
 +----------------------------------------------------------------------------*/
unsigned char level2dims ( unsigned char level )
{
   unsigned char dims, base;

   /* The value of 'base' appears to vary randomly, being either 2 or 3  */
   /* with approximately equal likelyhood.  The value 4 has occasionally */
   /* been observed.  We use 3 here.                                     */
   base = 3;
   dims = base + 11 * (level - 1);
   dims = max(0, min(210, dims));

   return dims;
} 
   
/*----------------------------------------------------------------------------+
 | Converts dims level (0-210) to (int) percent of full ON voltage.           |
 | Linearized from measured data on an X10 LM465 (1-way) Lamp Module.         | 
 +----------------------------------------------------------------------------*/ 
int dims2pct ( unsigned char level )
{
   long int pct;

   pct = (level > 32) ? 90 * (level - 33) / 177 + 10 :
         (level > 1)  ?  8 * (level -  2) / 31  +  2 : 0 ;

   return (int)pct;
}    

#define MEASURED_DATA
#ifdef MEASURED_DATA
/*----------------------------------------------------------------------------+
 | Converts old-style preset level (zero-based, 0-31) to (int) percent of     |
 | full ON voltage.                                                           |
 | This is based on measurement of output voltage for a LampLinc 2000STW      |
 | (2-way) dimmer module.                                                     | 
 +----------------------------------------------------------------------------*/
int presetlevel2pct ( unsigned char level )
{
   static unsigned char table[32] = {
       0, 18, 21, 23, 27, 28, 31, 34, 36, 39, 42, 45, 48, 51, 54, 57,
      60, 63, 67, 70, 73, 76, 79, 82, 85, 87, 90, 92, 95, 97, 99, 100
   };

   return (int)table[max(0, min(31, level))]; 
}

/*----------------------------------------------------------------------------+
 | Converts extended preset level (0-63) to (int) percent of full ON voltage. |
 | (This is based on measurement of output voltage for a LM14A module.)       | 
 +----------------------------------------------------------------------------*/ 
int ext3level2pct ( unsigned char level )
{
   static unsigned char table[64] = {
       0, 12, 13, 15, 15, 17, 18, 20, 21, 22, 
      23, 25, 26, 28, 29, 31, 32, 34, 35, 37,
      38, 40, 41, 43, 44, 46, 48, 50, 51, 53, 
      54, 56, 58, 60, 61, 63, 64, 66, 67, 69, 
      71, 73, 74, 76, 77, 79, 80, 82, 83, 85, 
      86, 88, 88, 90, 91, 92, 93, 95, 96, 97, 
      98, 99, 100, 101
   };

   return (int)table[max(0, min(63, level))];
}
#else
/*----------------------------------------------------------------------------+
 | Converts (old) preset level (0-31) to (int) percent of full ON voltage.    |
 | This is just a linear translation.                                         | 
 +----------------------------------------------------------------------------*/ 
int presetlevel2pct ( unsigned char level )
{
   return 100 * (int)level / 31;
} 
   
/*----------------------------------------------------------------------------+
 | Converts extended preset level (0-63) to (int) percent of full ON voltage. |
 | (This is linearized from actual data - max deviation about 4% of full On). | 
 +----------------------------------------------------------------------------*/ 
int ext3level2pct ( unsigned char level )
{
   level = min(62, level);
   return (level > 0) ? 12 + 88 * ((int)level - 1) / 61 : 0;
}
#endif /* Measured vs Linearized data */


/*----------------------------------------------------------------------------+
 | Return the name of the function for the argument trigger signal            |
 +----------------------------------------------------------------------------*/
char *lookup_function ( int signal ) 
{
   int j;

   for ( j = 0; j < NTRIGTYPES; j++ ) {
      if ( trig_type[j].signal == signal )
         return trig_type[j].label;
   }
   return "";
}

/*----------------------------------------------------------------------------+
 | Initialize the state structure elements for (X10 encoded) hcode            |
 +----------------------------------------------------------------------------*/
void x10state_init ( unsigned char hcode )
{
   int ucode;

   x10state[hcode].lastunit = 0;
   x10state[hcode].lastcmd = 0xff;
   x10state[hcode].reset = 0;
   x10state[hcode].addressed = 0;
   x10state[hcode].state[OnState] = 0;
   x10state[hcode].state[DimState] = 0;
   x10state[hcode].state[ChgState] = 0;
   x10state[hcode].state[LightsOnState] = 0;
   x10state[hcode].launched = 0;
   for ( ucode = 0; ucode < 16; ucode++ ) {
      x10state[hcode].dimlevel[ucode] = 0;
      x10state[hcode].memlevel[ucode] = 0;
   }
   return;
}

/*----------------------------------------------------------------------------+
 | Initialize the state structure for all housecodes.                         |
 +----------------------------------------------------------------------------*/
void x10state_init_all ( void )
{
   int hcode;

   for ( hcode = 0; hcode < 16; hcode++ ) 
      x10state_init(hcode);
   x10state[0].lasthc = 0xff;

   return;
}

   
/*----------------------------------------------------------------------------+
 | Send a control command to the heyu monitor/state engine                    |
 +----------------------------------------------------------------------------*/
void send_x10state_command ( unsigned char command, unsigned char parm )
{
   extern int sptty;
   static unsigned char lbuf[4] = {0xff, 0xff, 0xff, 3};
   static unsigned char buf[3] = {ST_COMMAND, 0, 0};

   buf[1] = command;
   buf[2] = parm;

   write(sptty, lbuf, 4);
   write(sptty, buf, 3);
   return;
}


/*----------------------------------------------------------------------------+
 | Check the dims table to determine the onstate and dimstate of modules on   |
 | argument housecode, returning X10 bitmaps back through the argument list.  |
 +----------------------------------------------------------------------------*/
void get_states ( unsigned char hcode, unsigned int *onstate, unsigned int *dimstate )
{
   int           ucode;
   unsigned char level;
   unsigned int  mask, offstate, extmask, premask, stdmask;

   extmask  = modmask[Ext3Mask][hcode];
   premask  = modmask[PresetMask][hcode];
   stdmask  = modmask[StdMask][hcode];

   offstate = 0;
   *dimstate = 0;
   for ( ucode = 0; ucode < 16; ucode++ ) {
      mask = (1 << ucode);
      level = x10state[hcode].dimlevel[ucode];
      if ( level == 0 )
         offstate |= mask;
      else if ( (mask & extmask && level < 62)  ||
                (mask & premask && level < 31)  ||
                (mask & stdmask && level < 210)   ) {
         *dimstate |= mask;
      }
   }
   *onstate = ~offstate;

   return;
}
      
/*----------------------------------------------------------------------------+
 | Determine the dimstate of modules on argument housecode, returning an X10  |
 | bitmap of units which are dimmed.                                          |
 +----------------------------------------------------------------------------*/
unsigned int get_dimstate ( unsigned char hcode )
{
   int ucode;
   unsigned int bitmap, mask, dimstate = 0;

   /* Check Extended Preset Dim Units */
   bitmap = modmask[Ext3DimMask][hcode];
   if ( bitmap ) {
      mask = 1;
      for ( ucode = 0; ucode < 16; ucode++ ) {
         if ( bitmap & mask &&
              x10state[hcode].dimlevel[ucode] > 0 &&
              x10state[hcode].dimlevel[ucode] < 62  ) {
            dimstate |= mask;
         }
         mask = mask << 1;
      }
   }

   /* Check Preset Units */
   bitmap = modmask[PresetMask][hcode];
   if ( bitmap ) {
      mask = 1;
      for ( ucode = 0; ucode < 16; ucode++ ) {
         if ( bitmap & mask &&
              x10state[hcode].dimlevel[ucode] > 0 &&
              x10state[hcode].dimlevel[ucode] < 31  ) {
            dimstate |= mask;
         }
         mask = mask << 1;
      }
   }

   /* Check Standard Non-Preset Units */
   bitmap = (modmask[DimMask][hcode] | modmask[BriMask][hcode]) &
            ~modmask[Ext3DimMask][hcode] & ~modmask[PresetMask][hcode];
   if ( bitmap ) {
      mask = 1;
      for ( ucode = 0; ucode < 16; ucode++ ) {
         if ( bitmap & mask &&
              x10state[hcode].dimlevel[ucode] > 0 &&
              x10state[hcode].dimlevel[ucode] < 210  ) {
            dimstate |= mask;
         }
         mask = mask << 1;
      }
   }

   return dimstate;
}


/*----------------------------------------------------------------------------+
 | Copy current dimlevel > 0 (0 is the same as OFF) to memlevel for bitmap    |
 | units.
 | This is only valid for units which support the preset or extended preset   |
 | commands.                                                                  |
 +----------------------------------------------------------------------------*/
void save_dimlevel ( unsigned char hcode, unsigned int bitmap )
{
   int ucode;
   int level;
   unsigned int mask = 1;

   for ( ucode = 0; ucode < 16; ucode++ ) {
      if ( bitmap & mask ) {
         level = x10state[hcode].dimlevel[ucode];
         if ( level > 0 )
           x10state[hcode].memlevel[ucode] = level;
      }
      mask = mask << 1;
   }
   return;
}

/*----------------------------------------------------------------------------+
 | Restore current dimlevel from memlevel for bitmap units                    |
 | This is only valid for units which support the preset or extended preset   |
 | commands.                                                                  |
 +----------------------------------------------------------------------------*/
void restore_dimlevel ( unsigned char hcode, unsigned int bitmap )
{
   int ucode;
   int level;
   unsigned int mask = 1;

   for ( ucode = 0; ucode < 16; ucode++ ) {
      if ( bitmap & mask ) {
         level = x10state[hcode].memlevel[ucode];
         x10state[hcode].dimlevel[ucode] = level;
      }
      mask = mask << 1;
   }
   return;
}


/*----------------------------------------------------------------------------+
 | Set a specific dimlevel.                                                   |
 +----------------------------------------------------------------------------*/
void set_dimlevel ( int level , unsigned char hcode, unsigned int bitmap )
{
   int ucode;
   unsigned int mask = 1;

   for ( ucode = 0; ucode < 16; ucode++ ) {
      if ( bitmap & mask ) {
         x10state[hcode].dimlevel[ucode] = level;
      }
      mask = mask << 1;
   }
   return;
}

   
/*----------------------------------------------------------------------------+
 | Set dimlevels to '0n' level.                                               |
 +----------------------------------------------------------------------------*/
void set_max_level ( unsigned char hcode, unsigned int bitmap )
{
   int           ucode;

   if ( !bitmap )
      return;

   for ( ucode = 0; ucode < 16; ucode++ ) {
      if ( bitmap & (1 << ucode) ) {
         x10state[hcode].dimlevel[ucode] = maxdimlevel[hcode][ucode];
      }
   }

   return;
}

/*----------------------------------------------------------------------------+
 | Set dimlevels to '0n' level.                                               |
 +----------------------------------------------------------------------------*/
void set_on_level ( unsigned char hcode, unsigned int bitmap )
{
   int           ucode;

   if ( !bitmap )
      return;

   for ( ucode = 0; ucode < 16; ucode++ ) {
      if ( bitmap & (1 << ucode) ) {
         x10state[hcode].dimlevel[ucode] = ondimlevel[hcode][ucode];
      }
   }

   return;
}

/*----------------------------------------------------------------------------+
 | Update the stored dimlevel by the amount of the argument level 0-210,      |
 | positive for brightens, negative for dims.                                 |
 +----------------------------------------------------------------------------*/
void update_dimlevel ( int level , unsigned char hcode, unsigned int bitmap )
{
   int          ucode, newlevel, abslev, table, delta;
   unsigned int active, mask;

   /* Table used in conversion of dims to change in ext3 preset level */
   /* Table[0] is used for Base-2 dims, table[1] for others */
   static int ext3table[2][17] = {
      {0, 5, 9, 14, 18, 22, 27, 31, 36, 40, 44, 49, 53, 58, 61, 62, 62},
      {0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 62},
   };

   /* Table used in conversion of dims to change in old-style Preset level */
   /* Table [0] is brights from level 0, [1] is dims from level 31.        */
   static int presettable[2][15] = {
      {1, 1, 3, 6, 8, 11, 14, 16, 19, 22, 24, 27, 29, 30, 32},
      {1, 1, 4, 7, 9, 12, 15, 17, 20, 22, 25, 28, 30, 32, 32},
   };

   if ( !bitmap )
      return;

   /* Standard Dimmable Units */
   active = bitmap & (modmask[DimMask][hcode] | modmask[BriMask][hcode]) &
                    ~(modmask[Ext3Mask][hcode] | modmask[PresetMask][hcode]);
   if ( active ) {
      mask = 1;
      for ( ucode = 0; ucode < 16; ucode++ ) {
         if ( active & mask ) {
            newlevel = level + x10state[hcode].dimlevel[ucode];
            newlevel = min(210, max(3, newlevel));
            x10state[hcode].dimlevel[ucode] = newlevel;
         }
         mask = mask << 1;
      }
   }

   /* Ext3 Units */
   active = bitmap & modmask[Ext3Mask][hcode];
   if ( active ) {
      abslev = abs(level);
      table = (abslev % 11 == 2) ? 0 : 1;
      delta = ext3table[table][min(abslev/11, 16)];
      delta = (level < 0) ? -delta : delta;
      mask = 1;
      for ( ucode = 0; ucode < 16; ucode++ ) {
         if ( active & mask ) {
            newlevel = delta + x10state[hcode].dimlevel[ucode];
            newlevel = min(62, max(1, newlevel));
            x10state[hcode].dimlevel[ucode] = newlevel;
         }
         mask = mask << 1;
      }
   }

   /* Preset1/Preset2 Units, zero-base levels 0-31 */
   active = bitmap & modmask[PresetMask][hcode];
   if ( active ) {
      abslev = abs(level);
      delta = (level > 0) ? presettable[0][min(abslev/11, 14)] :
                           -presettable[1][min(abslev/11, 14)];
      mask = 1;
      for ( ucode = 0; ucode < 16; ucode++ ) {
         if ( active & mask ) {
            newlevel = delta + x10state[hcode].dimlevel[ucode];
            newlevel = min(31, max(0, newlevel));
            x10state[hcode].dimlevel[ucode] = newlevel;
         }
         mask = mask << 1;
      }
   }

   return;
}      


/*----------------------------------------------------------------------------+
 | Update the addressed/unaddresssed state of housecode|units                 |
 | Addressing is cumulative until reset by an intervening function code with  |
 | the same housecode.                                                        |
 | An alloff (all_units_off) function unaddresses all units on that housecode.|
 +----------------------------------------------------------------------------*/
void x10state_update_addr ( unsigned char addrbyte )
{
   unsigned char hcode, ucode;
   unsigned int  bitmap;

   hcode = (addrbyte & 0xf0) >> 4;
   ucode = addrbyte & 0x0f;
   bitmap = 1 << ucode;

   x10state[0].lasthc = hcode;
   x10state[hcode].lastunit = code2unit(ucode);

   if ( x10state[hcode].reset ) {
      x10state[hcode].addressed = bitmap;
      x10state[hcode].reset = 0;
   }
   else
      x10state[hcode].addressed |= bitmap;

   return;
}

/*----------------------------------------------------------------------------+
 | Update the x10state structure per the contents of the argument 'buf'       |
 | buf[0] is the housecode|function byte, buf[1] is the dim level 1-210 for   |
 | dims and brights.  (This routine handles just the standard X10 functions - |
 | Type 3 Extended code functions are handled separately.)                    |
 |                                                                            |
 | If the user has defined the type of module, e.g. lamp module, for a        |
 | specific housecode|unit address, the change of state for that address      |
 | is filtered by the supported features of that type of module, e.g. a dim   |
 | signal will be ignored for appliance modules.                              |
 |                                                                            |
 | The received signal and state are tested to see if any of the conditions   |
 | exist for triggering the launching of an external script, and if so, the   |
 | index of the launcher is passed back through argument 'launchp',           | 
 | otherwise -1 is passed back.                                               |
 +----------------------------------------------------------------------------*/
void x10state_update_func ( unsigned char *buf, int *launchp )
{
   unsigned char hcode, func, level;
   unsigned char actfunc, genfunc;
   unsigned int  trigaddr, active, mask, trigactive;
   unsigned int  resumask, fullmask, fulloffmask, applmask, bbdimask, ext3dimask, lofmask;
   unsigned int  bmaplaunch;
   unsigned int  onstate, dimstate;
   unsigned long afuncmap, gfuncmap;
   int           j, dims;
   LAUNCHER      *launcherp;
   extern unsigned char signal_source;

   launcherp = config.launcherp;

   *launchp = -1;
  
   if ( !i_am_state )
      return;

   func = buf[0] & 0x0f;
   hcode = (buf[0] & 0xf0) >> 4;
   genfunc = actfunc = func;

   x10state[hcode].reset = 1;

   /* Preset1/Preset2 functions take special handling */
   if ( func == 10 || func == 11 ) {
      level = hcode;
      hcode = x10state[0].lasthc;
      x10state[0].lasthc = level;
      /* Use zero-base level (0-31) here */
      level = rev_low_nybble(level) + 16 * (func - 10);      
   }
   else {
      x10state[0].lasthc = hcode;
      level = 0;
   }

   gfuncmap = 0;
   mask = 0;
   x10state[hcode].state[ChgState] = 0;
   x10state[hcode].launched = 0;

   trigaddr = x10state[hcode].addressed;

   switch ( func ) {
      case  0 :  /* AllOff */
         /* state = OffState; */
         afuncmap = (1 << AllOffTrig);
         gfuncmap = (1 << OffTrig);
         genfunc = OffFunc;
         mask = modmask[AllOffMask][hcode];
         trigaddr = 0xffff;
         onstate = x10state[hcode].state[OnState];
         active = modmask[AllOffMask][hcode];
         resumask = modmask[ResumeMask][hcode] & modmask[Ext3DimMask][hcode];
         lofmask = modmask[PresetMask][hcode] &
                   modmask[LightsOnFullMask][hcode] &
                   x10state[hcode].state[LightsOnState]; 
         save_dimlevel(hcode, active & resumask & ~lofmask);
         set_dimlevel(0, hcode, active);
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));
         x10state[hcode].addressed = 0;
         x10state[hcode].state[OnState] = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case  1 :  /* LightsOn */
         /* state = OnState; */
         afuncmap = (1 << LightsOnTrig);
         gfuncmap = (1 << OnTrig);
         genfunc = OnFunc;
         mask = modmask[LightsOnMask][hcode];
         trigaddr = 0xffff;
         resumask = modmask[ResumeMask][hcode];
         fullmask = modmask[LightsOnFullMask][hcode];
         onstate = x10state[hcode].state[OnState];
         active = modmask[LightsOnMask][hcode];
         /* Units which resume saved brightness */
         restore_dimlevel(hcode, active & resumask);
         /* Units which go to 'On' level */
         set_on_level(hcode, active & ~resumask & ~fullmask & modmask[OnFullMask][hcode]); 
         /* Units which go to full brightness */
         set_max_level(hcode, active & 
           (fullmask | (~onstate & modmask[OnFullOffMask][hcode])));
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));

         x10state[hcode].state[OnState] = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] |= active;
         break;
      case  2 :  /* On */
         /* state = OnState; */
         afuncmap = gfuncmap = (1 << OnTrig);
         if ( x10state[hcode].addressed == 0xffff ) {
            afuncmap = (1 << AllOnTrig);
         }
         mask = modmask[OnMask][hcode];
         onstate = x10state[hcode].state[OnState];
         resumask = modmask[ResumeMask][hcode];
         fullmask = modmask[OnFullMask][hcode];
         fulloffmask = modmask[OnFullOffMask][hcode];
         applmask = ~(modmask[DimMask][hcode] | modmask[BriMask][hcode]);
         active = x10state[hcode].addressed & modmask[OnMask][hcode];
         /* Resume-enabled units will go to saved brightness */
         restore_dimlevel(hcode, active & resumask);
         /* Modules which go to 'On' brightness */
         set_on_level(hcode, active & ~resumask &
           ((~onstate & fulloffmask) | fullmask | applmask));
         save_dimlevel(hcode, active & resumask);
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));

         x10state[hcode].state[OnState] = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case  3 :  /* Off */
         /* state = OffState; */
         afuncmap = gfuncmap = (1 << OffTrig);
         mask = modmask[OffMask][hcode];
         onstate = x10state[hcode].state[OnState];
         active = x10state[hcode].addressed & modmask[OffMask][hcode];
         resumask = modmask[ResumeMask][hcode];
         lofmask = modmask[PresetMask][hcode] &
                   modmask[LightsOnFullMask][hcode] &
                   x10state[hcode].state[LightsOnState]; 
         save_dimlevel(hcode, active & resumask & ~lofmask);
         set_dimlevel(0, hcode, active);
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));

         x10state[hcode].state[OnState]  = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case  4 :  /* Dim */
         /* state = DimState; */
         afuncmap = gfuncmap = (1 << DimTrig);
         genfunc = DimFunc;
         mask = modmask[DimMask][hcode];
         onstate = x10state[hcode].state[OnState];
         dims = -(int)buf[1];
         active = x10state[hcode].addressed & modmask[DimMask][hcode];
         bbdimask = modmask[BriDimMask][hcode];
         resumask = modmask[ResumeMask][hcode];
         ext3dimask = modmask[Ext3DimMask][hcode];
         /* Off resume-enabled units will first go to saved brightness */
         restore_dimlevel(hcode, active & resumask & ~onstate & ext3dimask);
         /* Modules which first go to full brightness if Off */
         set_max_level(hcode, active & bbdimask & ~onstate);
         update_dimlevel(dims, hcode, active);
         lofmask = modmask[PresetMask][hcode] &
                   modmask[LightsOnFullMask][hcode] &
                   x10state[hcode].state[LightsOnState]; 
         save_dimlevel(hcode, active & resumask & ~lofmask);
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));
         x10state[hcode].state[OnState]  = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case  5 :  /* Bright */
         /* state = DimState; */
         afuncmap = (1 << BriTrig);
         gfuncmap = (1 << DimTrig);
         genfunc = DimFunc;
         mask = modmask[BriMask][hcode];
         dims = (int)buf[1];
         active = x10state[hcode].addressed & modmask[BriMask][hcode];
         onstate = x10state[hcode].state[OnState];
         bbdimask = modmask[BriDimMask][hcode];
         resumask = modmask[ResumeMask][hcode];
         /* Off, resume-enabled units will first go to saved brightness */
         restore_dimlevel(hcode, active & resumask & ~onstate);
         /* Modules which first go to full brightness if Off */
         set_max_level(hcode, active & bbdimask & ~onstate);
         update_dimlevel(dims, hcode, active);
         lofmask = modmask[PresetMask][hcode] &
                   modmask[LightsOnFullMask][hcode] &
                   x10state[hcode].state[LightsOnState]; 
         save_dimlevel(hcode, active & resumask & ~lofmask);
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));
         x10state[hcode].state[OnState]  = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case  6 : /* LightsOff */
         /* state = OffState; */
         afuncmap = (1 << LightsOffTrig);
         gfuncmap = (1 << OffTrig);
         genfunc = OffFunc;
         mask = modmask[LightsOffMask][hcode];
         trigaddr = 0xffff;
         onstate = x10state[hcode].state[OnState];
         active = modmask[LightsOffMask][hcode];
         resumask = modmask[ResumeMask][hcode];
         lofmask = modmask[PresetMask][hcode] &
                   modmask[LightsOnFullMask][hcode] &
                   x10state[hcode].state[LightsOnState]; 
         save_dimlevel(hcode, active & resumask & ~lofmask);
         set_dimlevel(0, hcode, active);
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));
         x10state[hcode].state[OnState]  = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case 10 :  /* Preset1 */
      case 11 :  /* Preset2 */
         if ( hcode == 0xff ) {
            /* Housecode unknown - can't do anything */
            return;
         }
         onstate = x10state[hcode].state[OnState];
         dimstate = x10state[hcode].state[DimState];
         mask = modmask[PresetMask][hcode];
         active = x10state[hcode].addressed & modmask[PresetMask][hcode];
         resumask = modmask[ResumeMask][hcode];
         afuncmap = (1 << PresetTrig);
         if ( level == 0 ) {
            /* Lowest level => Off */
            /* state = OffState; */
            gfuncmap = (1 << OffTrig);
            genfunc = OffFunc; 
            save_dimlevel(hcode, active & resumask);
            set_dimlevel(0, hcode, active);
            get_states(hcode, &onstate, &dimstate);
            x10state[hcode].state[ChgState] = active &
              ((x10state[hcode].state[OnState] ^ onstate) |
               (x10state[hcode].state[DimState] ^ dimstate));
            x10state[hcode].state[OnState]  = onstate;
            x10state[hcode].state[DimState] = dimstate;
         }
         else if ( level == 31 ) {
            /* Highest level => On */
            /* state = OnState; */
            gfuncmap = (1 << OnTrig);
            genfunc = OnFunc; 
            set_dimlevel(31, hcode, active);
            save_dimlevel(hcode, active & resumask);
            get_states(hcode, &onstate, &dimstate);
            x10state[hcode].state[ChgState] = active &
              ((x10state[hcode].state[OnState] ^ onstate) |
               (x10state[hcode].state[DimState] ^ dimstate));
            x10state[hcode].state[OnState]  = onstate;
            x10state[hcode].state[DimState] = dimstate;
         }       
         else {
            /* Intermediate level => Dim */
            /* state = DimState; */
            gfuncmap = (1 << DimTrig);
            genfunc = DimFunc; 
            set_dimlevel(level, hcode, active);
            save_dimlevel(hcode, active & resumask);
            get_states(hcode, &onstate, &dimstate);
            x10state[hcode].state[ChgState] = active &
              ((x10state[hcode].state[OnState] ^ onstate) |
               (x10state[hcode].state[DimState] ^ dimstate));
            x10state[hcode].state[OnState]  = onstate;
            x10state[hcode].state[DimState] = dimstate;
         }
         x10state[hcode].state[LightsOnState] &= ~active;

         /* Preset modules unaddress themselves after receiving a preset function */
         /* (otherwise they'd be affected by the next non-Preset function on the  */
         /* same housecode), but not if they've _sent_ the preset in response to  */
         /* a status request.                                                     */
         if ( !(x10state[hcode].lastcmd == StatusReqFunc && signal_source & RCVI) )
            x10state[hcode].addressed &= ~active;
       
         break;
      case 13 :  /* Status On */
         /* state = OnState; */
         afuncmap = (1 << StatusOnTrig);
         mask = modmask[StatusOnMask][hcode];
         active = x10state[hcode].addressed & modmask[StatusOnMask][hcode];
         onstate = x10state[hcode].state[OnState];
         set_on_level(hcode, active & ~onstate);
         get_states(hcode, &onstate, &dimstate);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));
         x10state[hcode].state[OnState]  = onstate;
         x10state[hcode].state[DimState] = dimstate;
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case 14 :  /* Status Off */
         /* state = OffState; */
         afuncmap = (1 << StatusOffTrig);
         gfuncmap = 0;
         mask = modmask[StatusOffMask][hcode];
         active = x10state[hcode].addressed & modmask[StatusOffMask][hcode];
         onstate = x10state[hcode].state[OnState];
         dimstate = x10state[hcode].state[DimState];
         x10state[hcode].state[OnState]  &= ~active;
         x10state[hcode].state[DimState] &= ~active;
         set_dimlevel(0, hcode, active);
         x10state[hcode].state[ChgState] = active &
           ((x10state[hcode].state[OnState] ^ onstate) |
            (x10state[hcode].state[DimState] ^ dimstate));
         x10state[hcode].state[LightsOnState] &= ~active;
         break;
      case 15 : /* StatusReq */
         afuncmap = (1 << StatusReqTrig);
         break;
      case 8 :  /* HailReq */
         afuncmap = (1 << HailReqTrig);
         break;
      case 9 :  /* Hail Ack */
         afuncmap = (1 << HailAckTrig);
         break;
      case 12 : /* Data Xfer */
         afuncmap = (1 << DataXferTrig);
         break;
      default :
         /* state = OffState; */
         afuncmap = 0;
         mask = 0;
         trigaddr = 0;
         break;
   }

   x10state[hcode].lastcmd = func;

   write_x10state_file();

   /* Heyuhelper, if applicable */
   if ( signal_source & RCVI && config.script_mode & HEYU_HELPER ) {
      launch_heyuhelper(hcode, trigaddr, func);
      return;
   }
   
   bmaplaunch = 0;
   j = 0;
   while ( launcherp && launcherp[j].line_no > 0 ) {

      if ( launcherp[j].hcode == hcode &&
           (launcherp[j].afuncmap & afuncmap || launcherp[j].gfuncmap & gfuncmap)  &&
           launcherp[j].source & signal_source ) {
         trigactive = trigaddr & (mask | launcherp[j].signal);
         if ( (bmaplaunch = (trigactive & launcherp[j].bmaptrig) & 
                   (x10state[hcode].state[ChgState] | ~launcherp[j].chgtrig)) ) {
            *launchp = j;
            launcherp[j].bmaplaunch = bmaplaunch;
            launcherp[j].actfunc = func;
            launcherp[j].genfunc = genfunc;
	    launcherp[j].level = level + 1;  /* preset 1-32 */
            break;
         }
      }
      j++;
   }
   x10state[hcode].launched = bmaplaunch;

   return;
}

/*----------------------------------------------------------------------------+
 | Update the x10state structure per the contents of the argument 'buf'       |
 | for type 3 extended codes.  'buf' will always contain 4 bytes.             |
 |                                                                            |
 | If the user has defined the type of module, e.g. lamp module, for a        |
 | specific housecode|unit address, the change of state for that address      |
 | is filtered by the supported features of that type of module, e.g. an      |
 | extended preset dim signal will be ignored for appliance modules.          |
 |                                                                            |
 | The received signal and state are tested to see if any of the conditions   |
 | exist for triggering the launching of an external script, and if so, the   |
 | index of the launcher is passed back through argument 'launchp',           | 
 | otherwise -1 is passed back.                                               |
 +----------------------------------------------------------------------------*/
void x10state_update_ext3func ( unsigned char *buf, int *launchp )
{
   unsigned char hcode, func, xfunc, ucode, subunit, level;
   unsigned char actfunc, genfunc;
   unsigned int  bitmap, trigaddr, mask, active, trigactive, ext3switch, onstate, dimstate;
   unsigned int  bmaplaunch;
   unsigned long afuncmap, gfuncmap;
   int           j;
   LAUNCHER      *launcherp;
   extern unsigned char signal_source;

   launcherp = config.launcherp;

   *launchp = -1;

   if ( !i_am_state )
      return;

   func = buf[0] & 0x0f;
   if ( func != 7 )
      return;
   actfunc = genfunc = func;
   afuncmap = (1 << ExtendedTrig);

   hcode = (buf[0] & 0xf0) >> 4;
   ucode = buf[1] & 0x0f;

   subunit = (buf[1] & 0xf0) >> 4;
   /* Subunit states are not supported */
   if ( subunit > 0 ) 
      return;
	      
   level = buf[2] & 0x3f;
   xfunc = buf[3];
   bitmap = 1 << ucode;
   trigaddr = bitmap;

   x10state[hcode].lastcmd = func;
   x10state[hcode].state[ChgState] = 0;
   x10state[hcode].lastunit = code2unit(ucode);
   x10state[0].lasthc = hcode;

   x10state[hcode].reset = 1;

   onstate = x10state[hcode].state[OnState];
   dimstate = x10state[hcode].state[DimState];
   ext3switch = bitmap & modmask[Ext3Mask][hcode] & ~modmask[Ext3DimMask][hcode];

   switch ( xfunc ) {
      case 0x31 :  /* Ext Preset */
         if ( level > 0 && ext3switch ) {
            gfuncmap = (1 << OnTrig);
            genfunc = OnFunc; 
            mask = modmask[Ext3Mask][hcode];
            active = bitmap & mask;
            x10state[hcode].state[ChgState] = ~x10state[hcode].state[OnState] & active;
            x10state[hcode].state[OnState] |= active;
            x10state[hcode].state[DimState] &= active;
            set_dimlevel(62, hcode, active);
            save_dimlevel(hcode, active);
         }
         else if ( level >= 62 ) {
            /* Full On is level 62 or 63 */
            gfuncmap = (1 << OnTrig);
            genfunc = OnFunc; /* On function */
            mask = modmask[Ext3DimMask][hcode];
            active = bitmap & mask;
            x10state[hcode].state[ChgState] = 
               (~x10state[hcode].state[OnState] | x10state[hcode].state[DimState]) & active;
            x10state[hcode].state[OnState] |= active;
            x10state[hcode].state[DimState] &= ~active;
            set_dimlevel(62, hcode, active);
            save_dimlevel(hcode, active);
         }
         else if ( level == 0 ) {
            /* Level 0 is Off */
            gfuncmap = (1 << OffTrig);
            genfunc = OffFunc;
            mask = modmask[Ext3Mask][hcode];
            active = bitmap & mask;
            x10state[hcode].state[ChgState] =
              (x10state[hcode].state[OnState] | x10state[hcode].state[DimState]) & active;
            x10state[hcode].state[OnState] &= ~active;
            x10state[hcode].state[DimState] &= ~active;
            save_dimlevel(hcode, active);
            set_dimlevel(0, hcode, active);
         }
         else {
            /* state = DimState; */
            gfuncmap = (1 << DimTrig);
            genfunc = DimFunc;
            mask = modmask[Ext3DimMask][hcode];
            active = bitmap & mask;
            x10state[hcode].state[ChgState] = active;
            x10state[hcode].state[OnState] |= active;
            x10state[hcode].state[DimState] |= active;
            set_dimlevel(level, hcode, active);
            save_dimlevel(hcode, active);
         }
         break;
      case 0x33 : /* Ext AllOn */
         gfuncmap = (1 << OnTrig);
         genfunc = OnFunc;
         mask = modmask[Ext3Mask][hcode];
         trigaddr = 0xffff;
         active = modmask[Ext3Mask][hcode];
         x10state[hcode].state[ChgState] =
            (~x10state[hcode].state[OnState] | x10state[hcode].state[DimState]) & active;
         x10state[hcode].state[OnState] |= active;
         x10state[hcode].state[DimState] &= ~active;
         /* Ext AllOn sets full brightness regardless of previous dim level */
         set_dimlevel(62, hcode, active);
         save_dimlevel(hcode, active);
         break;
      case 0x34 : /* Ext AllOff */
         gfuncmap = (1 << OffTrig);
         genfunc = OffFunc; 
         mask = modmask[Ext3Mask][hcode];
         trigaddr = 0xffff;
         active = modmask[Ext3Mask][hcode];
         x10state[hcode].state[ChgState] =
            (x10state[hcode].state[OnState] | x10state[hcode].state[DimState]) & active;
         x10state[hcode].state[OnState] &= ~active;
         x10state[hcode].state[DimState] &= ~active;
         /* Does not set the module "memory" level to 0 */
         save_dimlevel(hcode, active & modmask[Ext3DimMask][hcode]);
         set_dimlevel(0, hcode, active);
         break;
      case 0x38 : /* Ext Status Ack */
         gfuncmap = 0;
         if ( level > 0 && ext3switch ) {
            gfuncmap = (1 << OnTrig);
            genfunc = OnFunc;
            mask = modmask[Ext3Mask][hcode];
            active = bitmap & mask;
            x10state[hcode].state[ChgState] = ~x10state[hcode].state[OnState] & active;
            x10state[hcode].state[OnState] |= active;
            x10state[hcode].state[DimState] &= ~active;
            set_dimlevel(62, hcode, active);
         }
         else if ( level >= 62 ) {
            gfuncmap = (1 << OnTrig);
            genfunc = OnFunc;
            mask = modmask[Ext3DimMask][hcode];
            active = bitmap & mask;
            x10state[hcode].state[ChgState] = ~x10state[hcode].state[OnState] & active;
            x10state[hcode].state[OnState] |= active;
            x10state[hcode].state[DimState] &= ~active;
            set_dimlevel(62, hcode, active);
            save_dimlevel(hcode, active);
         }
         else if ( level == 0 ) {
            /* This does not reflect the module's level  */
            /* "memory", only it's current status as OFF */
            gfuncmap = (1 << OffTrig);
            genfunc = OffFunc; 
            mask = modmask[Ext3Mask][hcode];
            active = bitmap & mask;
            x10state[hcode].state[ChgState] = x10state[hcode].state[OnState] & active;
            x10state[hcode].state[OnState] &= ~active;
            x10state[hcode].state[DimState] &= ~active;
            /* This may be inaccurate, depending on module's history */ 
            save_dimlevel(hcode, active & modmask[Ext3DimMask][hcode]);
            set_dimlevel(0, hcode, active);
         }
         else {
            /* state = DimState; */
            mask = modmask[Ext3DimMask][hcode];
            gfuncmap = (1 << DimTrig);
            genfunc = DimFunc;
            active = bitmap & mask;
            x10state[hcode].state[ChgState] = ~x10state[hcode].state[DimState] & active;
            x10state[hcode].state[OnState] |= active;
            x10state[hcode].state[DimState] |= active;
            set_dimlevel(level, hcode, active);
            save_dimlevel(hcode, active);
         }
         break;
      default :
         /* state = OffState; */
         mask = 0;
         gfuncmap = 0;
         genfunc = 0;
         break;
   }

   write_x10state_file();

   /* Heyuhelper, if applicable */
   if ( signal_source & RCVI && config.script_mode & HEYU_HELPER ) {
      launch_heyuhelper(hcode, trigaddr, func);
      return;
   }

   bmaplaunch = 0;
   j = 0;
   while ( launcherp && launcherp[j].line_no > 0 ) {
      if ( launcherp[j].hcode == hcode &&
           (launcherp[j].afuncmap & afuncmap || launcherp[j].gfuncmap & gfuncmap) &&
           launcherp[j].source & signal_source ) {
         trigactive = trigaddr & (mask | launcherp[j].signal);
         if ( (bmaplaunch = (trigactive & launcherp[j].bmaptrig) & 
                   (x10state[hcode].state[ChgState] | ~launcherp[j].chgtrig)) ) {
            *launchp = j;
            launcherp[j].actfunc = actfunc;
            launcherp[j].genfunc = genfunc;
	    launcherp[j].xfunc = xfunc;
	    launcherp[j].level = level;
            launcherp[j].bmaplaunch = bmaplaunch;
            break;
         }
      }
      j++;
   }
   x10state[hcode].launched = bmaplaunch;

   return;
}

/*---------------------------------------------------------------------+
 | Display the detailed state of each unit on the argument X10-encoded |
 | housecode, i.e., Addressed, On, Dimmed, Changed.                    |
 +---------------------------------------------------------------------*/
void x10state_show ( unsigned char hcode ) 
{
   char lasthc;
   unsigned char lastcmd;
   unsigned int lastunitbmap;
   char minibuf[32];
   char *chrs = ".*";
   int  lw = 13;


   lastcmd = x10state[hcode].lastcmd;
   if ( lastcmd == 0xff )
      sprintf(minibuf, "_none_");
   else 
      sprintf(minibuf, funclabel[lastcmd]);

   lastunitbmap = (x10state[hcode].lastunit > 0) ?
       (1 << unit2code(x10state[hcode].lastunit)) : 0;
 
   lasthc = (x10state[0].lasthc == 0xff) ? '_' : code2hc(x10state[0].lasthc);

   printf("%*s %c\n", lw + 13, "Housecode", code2hc(hcode));
   printf("%*s  1..4...8.......16\n", lw, "Unit:");
   printf("%*s (%s)\n", lw,    "LastUnit", bmap2asc(lastunitbmap, ".u"));
   printf("%*s (%s)\n", lw,   "Addressed", bmap2asc(x10state[hcode].addressed, ".a"));
   printf("%*s (%s)\n", lw,          "On", bmap2asc(x10state[hcode].state[OnState], chrs));
   printf("%*s (%s)\n", lw,      "Dimmed", bmap2asc(x10state[hcode].state[DimState], ".x"));
   printf("%*s (%s)\n", lw,     "Changed", bmap2asc(x10state[hcode].state[ChgState], ".c"));
   printf("%*s (%s)\n", lw,    "Launched", bmap2asc(x10state[hcode].launched, chrs));
   printf("Last function this housecode: %s\n", minibuf);
   printf("Last Housecode = %c\n", lasthc);
   printf("\n");
   fflush(stdout);
   return;
}

/*---------------------------------------------------------------------+
 | Return a 16 character ASCII string displaying in ascending order    |
 | an X10 unit bitmap, i.e., char[0] -> unit 1, char[15] -> unit 16.   |
 | The argument chrs is a three-character string, the 1st character of |
 | represents 'unset' units and the 2nd character the 'set' bits in    |
 | bitmap1.  The 3rd character will overwrite the 2nd character for  ` |
 | set bits in bitmap2                                                 |
 +---------------------------------------------------------------------*/
char *bmap2asc2 ( unsigned int bitmap1, unsigned int bitmap2, char *chrs )
{
   int j;
   static char outbuf[17];

   for ( j = 0; j < 16; j++ ) {
      if ( bitmap1 & (1 << j) )
         outbuf[code2unit(j) - 1] = chrs[1];
      else
         outbuf[code2unit(j) - 1] = chrs[0];
   }
   for ( j = 0; j < 16; j++ ) {
      if ( bitmap2 & (1 << j) )
         outbuf[code2unit(j) - 1] = chrs[2];
   }
   outbuf[16] = '\0';
   return outbuf;
}

/*---------------------------------------------------------------------+
 | For each Unit in the X10 encoded hcode, display a map of the        |
 | signals which will launch a script.                                 |
 +---------------------------------------------------------------------*/
void show_launcher ( unsigned char hcode )
{
   unsigned int  bitmap1, bitmap2;
   unsigned int  signal, funcmap, afuncmap, gfuncmap, bmaptrig, chgtrig;
   unsigned char source, found = 0;
   char          *chrs = ".*C", *fmode;
   int           j, lw = 13;
   LAUNCHER      *launcherp;

   if ( (launcherp = config.launcherp) == NULL ) {
      printf("No Launchers have been defined in the configuration file.\n");
      return;
   }

   j = 0;
   while ( launcherp[j].line_no > 0 ) {
      if ( launcherp[j].hcode != hcode ) {
         j++;
         continue;
      }
      found = 1;   
      signal = launcherp[j].signal;
      afuncmap = launcherp[j].afuncmap;
      gfuncmap = launcherp[j].gfuncmap;
      funcmap = gfuncmap | afuncmap;
      bmaptrig = launcherp[j].bmaptrig;
      chgtrig = launcherp[j].chgtrig;
      source = launcherp[j].source;

      printf("%*s %c\n", lw + 13, "Housecode", code2hc(hcode));
      printf("%*s  1..4...8.......16\n", lw, "Unit:");

      if ( funcmap & (1 << OnTrig) ) { 
         bitmap1 = (modmask[OnMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[OnMask][hcode] | signal) & chgtrig;
         fmode = (gfuncmap & (1 << OnTrig)) ? "Generic" : "";
      }
      else {
         bitmap1 = bitmap2 = 0;
         fmode = "";
      }
      printf("%*s (%s)  %s\n", lw, "On", bmap2asc2(bitmap1, bitmap2, chrs), fmode);

      if ( funcmap & (1 << OffTrig) ) {
         bitmap1 = (modmask[OffMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[OffMask][hcode] | signal) & chgtrig;
         fmode = (gfuncmap & (1 << OffTrig)) ? "Generic" : "";
      }
      else {
         bitmap1 = bitmap2 = 0;
         fmode = "";
      }
      printf("%*s (%s)  %s\n", lw, "Off", bmap2asc2(bitmap1, bitmap2, chrs), fmode);

      if ( funcmap & (1 << DimTrig) ) {
         bitmap1 = (modmask[DimMask][hcode] | modmask[Ext3DimMask][hcode] |
             modmask[PresetMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[DimMask][hcode] | modmask[Ext3DimMask][hcode] |
             modmask[PresetMask][hcode] | signal) & chgtrig;
         fmode = (gfuncmap & (1 << DimTrig)) ? "Generic" : "";
      }
      else {
         bitmap1 = bitmap2 = 0;
         fmode = "";
      }
      printf("%*s (%s)  %s\n", lw, "Dim", bmap2asc2(bitmap1, bitmap2, chrs), fmode);

      if ( funcmap & (1 << BriTrig) ) {
         bitmap1 = (modmask[BriMask][hcode] | modmask[Ext3DimMask][hcode] |
             modmask[PresetMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[BriMask][hcode] | modmask[Ext3DimMask][hcode] |
             modmask[PresetMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "Bright", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << LightsOnTrig) ) {
         bitmap1 = (modmask[LightsOnMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[LightsOnMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "LightsOn", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << LightsOffTrig) ) {
         bitmap1 = (modmask[LightsOffMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[LightsOffMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "LightsOff", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << AllOnTrig) ) {
         bitmap1 = (modmask[AllOnMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[AllOnMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "AllOn", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << AllOffTrig) ) {
         bitmap1 = (modmask[AllOffMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[AllOffMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "AllOff", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << PresetTrig) ) {
         bitmap1 = (modmask[PresetMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[PresetMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "Preset", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << ExtendedTrig) ) {
         bitmap1 = (modmask[Ext3Mask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[Ext3Mask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "Extended", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << StatusOnTrig) ) {
         bitmap1 = (modmask[StatusOnMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[StatusOnMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "StatusOn", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << StatusOffTrig) ) {
         bitmap1 = (modmask[StatusOffMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[StatusOffMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "StatusOff", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << StatusReqTrig) ) {
         bitmap1 = (modmask[StatusMask][hcode] | signal) & bmaptrig;
         bitmap2 = (modmask[StatusMask][hcode] | signal) & chgtrig;
      }
      else
         bitmap1 = bitmap2 = 0;
      printf("%*s (%s)\n", lw, "Status", bmap2asc2(bitmap1, bitmap2, chrs));

      if ( funcmap & (1 << HailReqTrig) ) {
         bitmap1 = signal & bmaptrig;
         printf("%*s (%s)\n", lw, "Hail", bmap2asc(bitmap1, ".*"));
      }

      if ( funcmap & (1 << HailAckTrig) ) {
         bitmap1 = signal & bmaptrig;
         printf("%*s (%s)\n", lw, "HailAck", bmap2asc(bitmap1, ".*"));
      }

      if ( funcmap & (1 << DataXferTrig) ) {
         bitmap1 = signal & bmaptrig;
         printf("%*s (%s)\n", lw, "DataXfer", bmap2asc(bitmap1, ".*"));
      }

      printf("\n");
      printf("%*s: %s\n", lw, "Launch Mode", ((bmaptrig & signal) ? "Signal" : "Module"));
      printf("%*s: ", lw, "Launch Source");
      if ( source & SNDS )  printf("snds ");
      if ( source & SNDC )  printf("sndc ");
      if ( source & SNDM )  printf("sndm ");
      if ( source & SNDT )  printf("sndt ");
      if ( source & RCVT )  printf("rcvt ");
      if ( source & RCVI )  printf("rcvi ");
      printf("\n");
      printf("%*s: %s\n", lw, "Script label", launcherp[j].label);

      printf("\n");

      j++;
   }

   if ( !found )
      printf("No launchers have been specified for this housecode.\n");

   return;
}


/*---------------------------------------------------------------------+
 | Display the On/Off/Dimmed state for all housecodes and units.       |
 +---------------------------------------------------------------------*/
void show_housemap ( void ) 
{
   unsigned char hcode;
   char          hc;
   int           j, unit;
   char          outbuf[17];
   char          label[16];
   char          *chrs = "*x.";
   int           lw = 13;

   printf("%*s %s\n", lw, "", "* = On  x = Dimmed");
   printf("%*s  1..4...8.......16\n", lw, "Unit:");
   for ( hc = 'A'; hc <= 'P'; hc++ ) {
      hcode = hc2code(hc);

      for ( j = 0; j < 16; j++ ) {
         unit = code2unit(j) - 1;
         if ( x10state[hcode].state[DimState] & (1 << j) ) {
            outbuf[unit] = chrs[1];
         }
         else if ( x10state[hcode].state[OnState] & (1 << j) ) {
            outbuf[unit] = chrs[0];
         }
         else
            outbuf[unit] = chrs[2];
      }
      sprintf(label, "Housecode %c", hc);
      outbuf[16] = '\0';
      printf("%*s (%s)\n", lw, label, outbuf);
   }
   return;
}

/*---------------------------------------------------------------------+
 | Function to delete the X10state file.                               |
 +---------------------------------------------------------------------*/
void remove_x10state_file ( void )
{
   int   code;

   if ( verbose )
      printf("Deleting x10state file %s\n", statefile);

   code = remove( statefile ) ;
   if ( code != 0 && errno != 2 ) {
      (void)fprintf(stderr, 
         "WARNING: Unable to delete X10 State File %s - errno = %d\n",
          statefile, errno);
   }
   return;
}

/*---------------------------------------------------------------------+
 | Function to write the x10state structure to the X10state file.      |
 +---------------------------------------------------------------------*/
void write_x10state_file ( void )
{
   FILE *fd;

   if ( !(fd = fopen(statefile, "w")) ) {
      fprintf(stderr, "Unable to open X10 State File '%s' for writing.\n",
         statefile);
      exit(1);
   }

   fwrite((void *)x10state, 1, (sizeof(x10state)), fd);
   fclose(fd);
   return;
}

/*---------------------------------------------------------------------+
 | Get the last-modified time of the x10state file.                    |
 +---------------------------------------------------------------------*/
long int modtime_x10state_file ( void )
{
   struct stat statbuf;
   int retcode;

   retcode = stat(statefile, &statbuf);

   if ( retcode == 0 )
      return statbuf.st_mtime;

   return (long)retcode;
}

/*---------------------------------------------------------------------+
 | Function to read the x10state file and store the contents in the    |
 | x10state structure.                                                 |
 +---------------------------------------------------------------------*/
int read_x10state_file (void )
{
   FILE *fd = NULL;  /* Keep some compilers happy; ditto with nread */
   int  j, nread = 0;
 
   for ( j = 0; j < 3; j++ ) {
      if ( (fd = fopen(statefile, "r")) )
         break;
      millisleep(100);
   }
   if ( !fd ) {
      return 1;
   }

   for ( j = 0; j < 3; j++ ) {
      nread = fread((void *)x10state, 1, (sizeof(x10state)), fd );
      if ( nread == (sizeof(x10state)) )
         break;
      rewind(fd);
      millisleep(100);
   }

   if ( nread != (sizeof(x10state)) ) {
      return 1;
   }
   fclose(fd);
   return 0;
}

/*---------------------------------------------------------------------+
 | Send command to monitor process to write the x10 state file, then   |
 | read it and store in the state structure.                           |
 +---------------------------------------------------------------------*/
int fetch_x10state ( void )
{
   struct stat statbuf;
   int         j, retcode;
   time_t      now;

   time(&now);
   send_x10state_command(ST_WRITE, 0);
   for ( j = 0; j < 5; j++ ) {
      if ( (retcode = stat(statefile, &statbuf)) == 0 &&
           statbuf.st_mtime >= now ) {
         if ( read_x10state_file() == 0 )
            break;
      }
      millisleep(100);
   }
   if ( j >= 5 ) {
      fprintf(stderr, "Unable to read current x10state - is state engine running?\n");
      return 1;
   }

   return 0;
}   

/*---------------------------------------------------------------------+
 | Identify the type of data read from the spoolfile which was written |
 | there by heyu, i.e., excluding data received from the interface.    |
 +---------------------------------------------------------------------*/
int identify_sent ( unsigned char *buf, int len )
{
   int type;

   if ( buf[0] == 0 && len == 1 )
      type = SENT_WRMI;  /* WRMI */
   else if ( buf[0] == 0x04 && len == 2 )
      type = SENT_ADDR;  /* Address */
   else if ( (buf[0] & 0x07) == 0x06 && len == 2 )
      type = SENT_FUNC;  /* Standard Function */
   else if ( buf[0] == 0x07 && len == 5 )
      type = SENT_EXTFUNC;  /* Extended function */
   else if ( buf[0] == ST_COMMAND && len == 3 )
      type = SENT_STCMD;  /* Monitor/state control command */
   else
      type = SENT_OTHER;  /* Other, i.e., info, setclock, upload, etc. */

   return type;
}

/*---------------------------------------------------------------------+
 | Interpret byte string sent to CM11A, update the state, and test     |
 | whether any launch condition is satisfied.                          |
 +---------------------------------------------------------------------*/
char *translate_sent ( unsigned char *buf, int len, int *launchp )
{ 
   static char outbuf[80];
   char hc;
   unsigned char hcode, func, xfunc, xdata, level, unit, subunit, chksum;
   unsigned char newbuf[2];
   unsigned int bmap, memloc;
   extern int checksum(unsigned char *, int);
   int dims;
   double ddims;

   *launchp = -1;

   if ( buf[0] == 0 && len == 1 ) {
      return "";
   }
   else if ( buf[0] == 0x04 && len == 2 ) {
      /* Address */
      hcode = (buf[1] & 0xf0) >> 4;
      hc = code2hc(hcode);
      unit = code2unit(buf[1] & 0x0f);
      bmap = 1 << (buf[1] & 0x0f);
      x10state_update_addr(buf[1]); 
      sprintf(outbuf, "address unit    %3d : housecode %c (%s)",
         unit, hc, alias_rev_lookup(hc, bmap));
      return outbuf;
   }
   else if ( (buf[0] & 0x07) == 0x06 && len == 2 ) {
      /* Standard function */
      level = (buf[0] & 0xf8) >> 3;
      dims = level2dims(level);
      ddims = 100. * (double)dims / 210.;
      hcode = (buf[1] & 0xf0) >> 4;
      hc = code2hc(hcode);
      func = buf[1] & 0x0f;
      if ( level > 0 && func != 4 && func != 5 ) {
         sprintf(outbuf, "Unknown transmission: %02x %02x", buf[0], buf[1]);
         return outbuf;
      }
      newbuf[0] = buf[1];
      newbuf[1] = dims;

      x10state_update_func(newbuf, launchp);

      switch ( func ) {
         case 4 :  /* Dim */
            sprintf(outbuf, "function %10s : housecode %c, dimmed by %%%02.0f [%d]",
               funclabel[func], hc, ddims, dims);
            return outbuf;
         case 5 :  /* Bright */
            sprintf(outbuf, "function %10s : housecode %c, brighten %%%02.0f [%d]",
               funclabel[func], hc, ddims, dims);
            return outbuf;
         case 10 : /* Preset level */
         case 11 : /* Preset level */
            level = rev_low_nybble(hcode) + 1;
            level += (func == 11) ? 16 : 0;
            sprintf(outbuf, "function %10s : level %d", funclabel[func], level);
            return outbuf;
         default:
            sprintf(outbuf, "function %10s : housecode %c", funclabel[func], hc);
      }
   }
   else if ( buf[0] == 0x07 && len == 5 ) {
      /* Extended code function */
      char stmp[16];
      if ( (buf[1] & 0x0f) != 0x07 ) {
         sprintf(outbuf, "Unknown transmission: %02x %02x %02x %02x %02x",
             buf[0], buf[1], buf[2], buf[3], buf[4]);
         return outbuf;
      }
      hcode = (buf[1] & 0xf0) >> 4;
      hc = code2hc(hcode);
      unit = code2unit(buf[2] & 0x0f);
      subunit = (buf[2] & 0xf0) >> 4;
      bmap = 1 << (buf[2] & 0x0f);
      xfunc = buf[4];
      xdata = buf[3];

      x10state_update_ext3func(buf + 1, launchp);

      if ( subunit == 0 )
         stmp[0] = '\0';
      else
	 sprintf(stmp, " subunit %02d", subunit);

      if ( xfunc == 0x31 )
         sprintf(outbuf, "function    xPreset : housecode %c unit %d%s level %d (%s)",
            hc, unit, stmp, xdata, alias_rev_lookup(hc, bmap));
      else if ( xfunc == 0x33 )
         sprintf(outbuf, "function     xAllOn  : housecode %c", hc);
      else if ( xfunc == 0x34 )
         sprintf(outbuf, "function    xAllOff : housecode %c", hc);
      else if ( xfunc == 0x37 ) {
        if ( (xdata & 0x30) == 0 ) 
           sprintf(outbuf, "function xStatusReq : housecode %c unit %d%s (%s)",
              hc, unit, stmp, alias_rev_lookup(hc, bmap));
        else if ( (xdata & 0x30) == 0x10 )
           sprintf(outbuf, "function    xStatus : POWER_UP housecode %c unit %d%s (%s)",
               hc, unit, stmp, alias_rev_lookup(hc, bmap));
        else  
           sprintf(outbuf, "function    xStatus : OTHER housecode %c unit %d%s data=0x%02x (%s)",
              hc, unit, stmp, xdata, alias_rev_lookup(hc, bmap));
      }
      else if ( xfunc == 0x3b ) {
           sprintf(outbuf, "function    xConfig : housecode %c unit %d%s mode=%d (%s)",
              hc, unit, stmp, xdata, alias_rev_lookup(hc, bmap));
      }
      else if ( xfunc == 0x38 ) {
         if ( xdata & 0x40 )
            sprintf(outbuf, "function    xStatus : Switch %c%d%s is %-3s %s (%s)",
               hc, unit, stmp, ((xdata & 0x3f) ? "On " : "Off"),
               ((xdata & 0x80) ? "LoadOK" : "NoLoad"), 
               alias_rev_lookup(hc, bmap));
         else
            sprintf(outbuf, "function    xStatus : Lamp %c%d%s at level %d, %s (%s)",
               hc, unit, stmp, xdata & 0x3f, 
               ((xdata & 0x80) ? "BulbOK" : "NoBulb"), 
               alias_rev_lookup(hc, bmap));
      }
      else if ( xfunc == 0xff )
            sprintf(outbuf, "function    ExtCode : Incomplete xcode in buffer.");
      else {
            sprintf(outbuf, "function   xFunc %02x : housecode %c unit %d%s data=0x%02x (%s)",
            xfunc, hc, unit, stmp, xdata, alias_rev_lookup(hc, bmap));
      }

      return outbuf;
   }
   else if ( buf[0] == 0xfb && len == 19 ) {
      chksum = checksum(buf + 1, len - 1);
      memloc = (buf[1] << 8) + buf[2];
      sprintf(outbuf, "Upload to EEPROM location %03x, checksum = %02x",
          memloc, chksum);
      return outbuf;
   }
   else if ( buf[0] == 0x8b && len == 1 ) {
      sprintf(outbuf, "Request CM11a info");
      return outbuf;
   }
   else if ( buf[0] == 0x9b && len == 7 ) {
      sprintf(outbuf, "Update CM11a info");
      return outbuf;
   }
   else if ( buf[0] == 0xeb && len == 1 ) {
      sprintf(outbuf, "Enable serial ring signal");
      return outbuf;
   }
   else if ( buf[0] == 0xdb && len == 1 ) {
      sprintf(outbuf, "Disable serial ring signal");
      return outbuf;
   }
   else if ( buf[0] == ST_COMMAND && len == 3 ) {
      sprintf(outbuf, "State command");
      return outbuf;
   }
   else {
      sprintf(outbuf, "Unknown transmission: %d bytes, 1st byte = %02x",
         len, buf[0]);
   }
   return outbuf;
}

/*---------------------------------------------------------------------+
 | Interpret byte string sent to CM11A.                                |
 +---------------------------------------------------------------------*/
char *translate_other ( unsigned char *buf, int len )
{ 
   static char outbuf[80];
   unsigned char chksum;
   unsigned int memloc;
   extern int checksum(unsigned char *, int);

   if ( buf[0] == 0 && len == 1 ) {
      return "";
   }
   else if ( buf[0] == 0xfb && len == 19 ) {
      chksum = checksum(buf + 1, len - 1);
      memloc = (buf[1] << 8) + buf[2];
      sprintf(outbuf, "Upload to EEPROM location %03x, checksum = %02x",
          memloc, chksum);
      return outbuf;
   }
   else if ( buf[0] == 0x8b && len == 1 ) {
      sprintf(outbuf, "Request CM11a info");
      return outbuf;
   }
   else if ( buf[0] == 0x9b && len == 7 ) {
      sprintf(outbuf, "Update CM11a info");
      return outbuf;
   }
   else if ( buf[0] == 0xeb && len == 1 ) {
      sprintf(outbuf, "Enable serial ring signal");
      return outbuf;
   }
   else if ( buf[0] == 0xdb && len == 1 ) {
      sprintf(outbuf, "Disable serial ring signal");
      return outbuf;
   }
   else if ( buf[0] == ST_COMMAND && len == 3 ) {
      sprintf(outbuf, "State command");
      return outbuf;
   }
   else {
      sprintf(outbuf, "Unknown transmission: %d bytes, 1st byte = %02x",
         len, buf[0]);
   }
   return outbuf;
}

/*---------------------------------------------------------------------+
 | Create argument list string for heyuhelper.                         |
 | If argument big = 0, a token for only the last unit is included,    |
 | otherwise tokens are include for each unit in the launcher bitmap.  |
 +---------------------------------------------------------------------*/
char *helper_string ( int index, int big )
{
   int           ucode;
   unsigned int  bitmap, mask;
   unsigned char actfunc;
   char          hc;
   LAUNCHER      *launcherp;
   char          buffer[256];
   char          minibuf[32];
   char          *dup;

   launcherp = config.launcherp;
   bitmap = launcherp[index].bmaplaunch;
   actfunc = launcherp[index].actfunc;
   hc = tolower(code2hc(launcherp[index].hcode));

   strcpy(buffer, ((big) ? "bighelper=" : "helper="));
   for ( ucode = 15; ucode >= 0; ucode-- ) {
      mask = 1 << ucode;
      if ( bitmap & mask ) {
         sprintf(minibuf, "%c%d%s ",
            hc, code2unit(ucode), funclabel[actfunc]);
         strcat(buffer, minibuf);
         if ( !big )
            break;
      }
   }
   if ( (dup = strdup(buffer)) == NULL ) {
      fprintf(stderr, "Out of memory in helper_string()\n");
      exit(1);
   }
   return dup;
}
 

/*---------------------------------------------------------------------+
 | Return the heyu state in the same bitmap format as is put into the  |
 | environment when a script is launched.                              |
 +---------------------------------------------------------------------*/
int get_heyu_state ( unsigned char hcode, unsigned char ucode ) 
{
   unsigned int  bitmap, value;
   int           level;

   bitmap = (1 << ucode);
   
   /* Start with the dim level (0-100) */
   level = x10state[hcode].dimlevel[ucode];
   if ( modmask[Ext3Mask][hcode] & bitmap )
      value = ext3level2pct(level);
   else if ( modmask[PresetMask][hcode] & bitmap )
      value = presetlevel2pct(level);
   else 
      value = dims2pct(level);

   /* Add the state bits */
   if ( bitmap & ~(modmask[DimMask][hcode] | modmask[BriMask][hcode]) )
      value |= HEYUMAP_APPL;         
   if ( bitmap & ~x10state[hcode].state[OnState] )
      value |= HEYUMAP_OFF;
   if ( bitmap & x10state[hcode].addressed )
      value |= HEYUMAP_ADDR;
   if ( bitmap & x10state[hcode].state[ChgState] )
      value |= HEYUMAP_CHG;
   if ( bitmap & x10state[hcode].state[DimState] )
      value |= HEYUMAP_DIM;
   if ( bitmap & x10state[hcode].state[OnState] )
      value |= HEYUMAP_ON;

   return value;
}
  
/*---------------------------------------------------------------------+
 | Return the heyu state in the same bitmap format as is put into the  |
 | environment when a script is launched with "-rawlevel" option.      |
 +---------------------------------------------------------------------*/
int get_heyu_rawstate ( unsigned char hcode, unsigned char ucode ) 
{
   unsigned int  bitmap, value;
   int           level;

   bitmap = (1 << ucode);
   
   /* Start with the raw dim level */
   level = x10state[hcode].dimlevel[ucode];
   if ( modmask[PresetMask][hcode] & bitmap )
      value = level + 1;
   else 
      value = level;

   /* Add the state bits */
   if ( bitmap & ~(modmask[DimMask][hcode] | modmask[BriMask][hcode]) )
      value |= HEYUMAP_APPL;         
   if ( bitmap & ~x10state[hcode].state[OnState] )
      value |= HEYUMAP_OFF;
   if ( bitmap & x10state[hcode].addressed )
      value |= HEYUMAP_ADDR;
   if ( bitmap & x10state[hcode].state[ChgState] )
      value |= HEYUMAP_CHG;
   if ( bitmap & x10state[hcode].state[DimState] )
      value |= HEYUMAP_DIM;
   if ( bitmap & x10state[hcode].state[OnState] )
      value |= HEYUMAP_ON;

   return value;
}
  
/*---------------------------------------------------------------------+
 | Return the xtend state in the same bitmap format as is put into the |
 | environment when a script is launched.                              |
 +---------------------------------------------------------------------*/
int get_xtend_state ( unsigned char hcode, unsigned char ucode ) 
{
   unsigned int  bitmap, value;

   bitmap = (1 << ucode);

   value = 0;
   
   /* Add the state bits */
   if ( bitmap & ~(modmask[DimMask][hcode] | modmask[BriMask][hcode]) )
      value |= XTMAP_APPL;         
   if ( bitmap & x10state[hcode].addressed )
      value |= XTMAP_ADDR;
   if ( bitmap & x10state[hcode].state[OnState] )
      value |= XTMAP_ON;

   return value;
}
  
/*---------------------------------------------------------------------+
 | Allocate memory for argument envstr.                                |
 +---------------------------------------------------------------------*/
char *add_envptr ( char *envstr )
{
   char *dup;
   char *out_of_memory = "add_envptr() : Out of memory\n";
      
   
   if ( (dup = strdup(envstr)) == NULL ) {
      fprintf(stderr, out_of_memory);
      exit(1);
   }
   return dup;   
}
   
/*---------------------------------------------------------------------+
 | Create a minimal environment for the -noenv option                  |
 +---------------------------------------------------------------------*/
char **create_noenv_environment ( void )
{
   
   extern char   **environ;

   int           j, size, newsize;
   char          **envp, **ep;
   char          minibuf[512];
   static int    sizchrptr = sizeof(char *);

   /* Get length of original environment */
   size = 0;
   while ( environ[size] != NULL )
      size++;

   newsize = size + 50;

   if ( (envp = calloc(newsize, sizchrptr)) == NULL ) {
      fprintf(stderr, "create_noenv__environment() : out_of_memory\n");
      exit(1);
   }

   ep = envp;

   *ep++ = add_envptr("HEYU_MONITOR_CHILD=1");

   *ep++ = add_envptr("IAMNOENV=1");

   /* Put config file pathname in environment for child processes */
   sprintf(minibuf, "X10CONFIG=%s", pathspec(heyu_config));
   *ep++ = add_envptr(minibuf);


   /* Append the user's original environment */
   for ( j = 0; j < size; j++ ) {
      *ep++ = add_envptr(environ[j]);
   }

   /* Add NULL terminator */
   *ep++ = NULL;

   return envp;
}

/*---------------------------------------------------------------------+
 | Create the environment for heyu scripts.                            |
 +---------------------------------------------------------------------*/
char **create_heyu_environment( int index, unsigned char option )
{
   extern char   **environ;

   static char   *month[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	                     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
   static char   *wday[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
   
   LAUNCHER      *launcherp;
   int           j, size, newsize, level;
   unsigned char hcode, ucode, actfunc;
   char          **envp, **ep;
   char          hc;
   unsigned int  value, bitmap, addressed, onstate, appliance;
   unsigned int  dimstate, chgstate;
   long int      dawn, dusk, systime;
   int           dark, expire;
   char          *aliaslabel;
   char          minibuf[512];
   static int    sizchrptr = sizeof(char *);
   time_t        now;
   struct tm     *tms;

   static struct {
      char *query;
      int  mask;
   } maskval[] = {
      {"whatLevel", HEYUMAP_LEVEL},
      {"isAppl",    HEYUMAP_APPL },
      {"isOff",     HEYUMAP_OFF  },
      {"isAddr",    HEYUMAP_ADDR },
      {"isChg",     HEYUMAP_CHG  },
      {"isDim",     HEYUMAP_DIM  },
      {"isOn",      HEYUMAP_ON   },
      { NULL,       0            },
   };

   
   /* Get length of original environment */
   size = 0;
   while ( environ[size] != NULL )
      size++;

   newsize = size + 256 + 50 + ((option & ENV_ALIAS) ? alias_count() : 0);

   if ( (envp = calloc(newsize, sizchrptr)) == NULL ) {
      fprintf(stderr, "create_heyu_environment() : out of memory\n");
      exit(1);
   }

   ep = envp;

   *ep++ = add_envptr("HEYU_MONITOR_CHILD=1");

   *ep++ = add_envptr("IAMHEYU=1");

   /* Put config file pathname in environment for child processes */
   sprintf(minibuf, "X10CONFIG=%s", pathspec(heyu_config));
   *ep++ = add_envptr(minibuf);

   /* Create the various masks */
   j = 0;
   while ( maskval[j].query ) {
      sprintf(minibuf, "%s=%d", maskval[j].query, maskval[j].mask);
      *ep++ = add_envptr(minibuf);
      j++;
   }

   /* Put in the variables 'helper' and 'bighelper' */
   /* with arguments for heyuhelper                 */
   *ep++ = helper_string(index, 0);
   *ep++ = helper_string(index, 1);

   launcherp = config.launcherp;
   hcode = launcherp[index].hcode;
   actfunc = launcherp[index].actfunc;
   
   sprintf(minibuf, "X10_Function=%s", funclabel[actfunc]);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "X10_Housecode=%c", code2hc(hcode));
   *ep++ = add_envptr(minibuf);

   for ( ucode = 0; ucode < 16; ucode++ ) {
      if ( launcherp[index].bmaplaunch & (1 << ucode) )
         break;
   } 
   sprintf(minibuf, "X10_Unit=%d", code2unit(ucode));
   *ep++ = add_envptr(minibuf);

   sprintf(minibuf, "X10_LastUnit=%d", x10state[hcode].lastunit);
   *ep++ = add_envptr(minibuf);

   if ( actfunc == 10 || actfunc == 11 ) {
      sprintf(minibuf, "X10_PresetLevel=%d", launcherp[index].level);   
      *ep++ = add_envptr(minibuf);
   }	 
   
   if ( actfunc == 7 ) {
      sprintf(minibuf, "X10_Xfunc=%d", launcherp[index].xfunc);   
      *ep++ = add_envptr(minibuf);

      sprintf(minibuf, "X10_Xdata=%d", launcherp[index].level);   
      *ep++ = add_envptr(minibuf);
   }	 
   

   /* Put in the variables X10_DawnTime, X10_DuskTime, and X10_SysTime */
   dawn = dawndusk_today(DAWN_EVENT);
   if ( dawn > 0 ) {
      sprintf(minibuf, "%s=%ld", "X10_DawnTime", dawn);
      *ep++ = add_envptr(minibuf);
   }
   
   dusk = dawndusk_today(DUSK_EVENT);
   if ( dusk > 0 ) {
      sprintf(minibuf, "%s=%ld", "X10_DuskTime", dusk);
      *ep++ = add_envptr(minibuf);
   }

   /* Add Clock/Calendar variables */

   time(&now);
   tms = localtime(&now);

   systime = 3600L * (long)tms->tm_hour
	   + (long)(60 * tms->tm_min + tms->tm_sec);
   sprintf(minibuf, "%s=%ld", "X10_SysTime", systime);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_Year", tms->tm_year + 1900);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_Month", tms->tm_mon + 1);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%s", "X10_MonthName", month[tms->tm_mon]);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_Day", tms->tm_mday);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_WeekDay", tms->tm_wday);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%s", "X10_WeekDayName", wday[tms->tm_wday]);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_Hour", tms->tm_hour);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_Minute", tms->tm_min);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_Second", tms->tm_sec);
   *ep++ = add_envptr(minibuf);
   
   sprintf(minibuf, "%s=%d", "X10_isDST", ((tms->tm_isdst > 0) ? 1 : 0));
   *ep++ = add_envptr(minibuf);
   
   if ( dawn > 0 && dusk > 0 ) {
      /* Logical variable is true between dusk and dawn */
      if ( systime < dawn || systime > dusk )
         dark = 1;
      else
         dark = 0;
      sprintf(minibuf, "%s=%d", "X10_isNightTime", dark);
      *ep++ = add_envptr(minibuf);

      /* Logical variable is true in the interval after Dusk */
      /* and before Dawn by the number of minutes defined by */
      /* the config directive ISDARK_OFFSET                  */

      if ( systime < (dawn - (long)(60 * config.isdark_offset)) ||
           systime > (dusk + (long)(60 * config.isdark_offset))    ) 
         dark = 1;
      else
         dark = 0;
      sprintf(minibuf, "%s=%d", "X10_isDarkTime", dark);
      *ep++ = add_envptr(minibuf);
   }

   /* Add upload expiration status and time */
   expire = get_upload_expire();

   sprintf(minibuf, "%s=%d", "X10_EXPIRE", expire);
   *ep++ = add_envptr(minibuf);

      
   /* Add bitmap for each Housecode|Unit */
   for ( hcode = 0; hcode < 16; hcode++ ) {
      hc = code2hc(hcode);
      addressed = x10state[hcode].addressed;
      onstate = x10state[hcode].state[OnState];
      dimstate = x10state[hcode].state[DimState];
      chgstate = x10state[hcode].state[ChgState];
      appliance = ~(modmask[DimMask][hcode] | modmask[BriMask][hcode]);
      for ( ucode = 0; ucode < 16; ucode++ ) {
         bitmap = 1 << ucode;

         /* Start with the dim level either raw or 0-100% */
         level = x10state[hcode].dimlevel[ucode];
         if ( option & LEVEL_RAW ) {
            if ( modmask[PresetMask][hcode] & bitmap )
               value = level + 1;
            else
               value = level;
         }
         else {
            if ( modmask[Ext3Mask][hcode] & bitmap )
               value = ext3level2pct(level);
            else if ( modmask[PresetMask][hcode] & bitmap )
               value = presetlevel2pct(level);
            else 
               value = dims2pct(level);
         }

         /* Add the state bits */
         if ( appliance & bitmap )
            value |= HEYUMAP_APPL;         
         if ( ~onstate & bitmap )
            value |= HEYUMAP_OFF;
         if ( addressed & bitmap )
            value |= HEYUMAP_ADDR;
         if ( chgstate & bitmap )
            value |= HEYUMAP_CHG;
         if ( dimstate & bitmap )
            value |= HEYUMAP_DIM;
         if ( onstate & bitmap )
            value |= HEYUMAP_ON;

         /* Add standard variable, e.g., X10_A9=nnnn */            
         sprintf(minibuf, "X10_%c%d=%d", hc, code2unit(ucode), value);
         *ep++ = add_envptr(minibuf);


         /* Add 'alias' variable, e.g., x10_porch_light=nnnn */
	 if ( option & ENV_ALIAS ) {
            j = 0;
	    while ( (aliaslabel = alias_rev_lookup_mult(hc, bitmap, &j)) != NULL ) {
               sprintf(minibuf, "%s_%s=%d",
		                config.env_alias_prefix, aliaslabel, value);
               *ep++ = add_envptr(minibuf);
            }
	 }
      }
   }

   /* Append the user's original environment */
   for ( j = 0; j < size; j++ ) {
      *ep++ = add_envptr(environ[j]);
   }

   /* Add NULL terminator */
   *ep++ = NULL;

   /* Make sure we haven't exceeded allocated size */
   if ( (int)(ep - envp) > newsize ) {
      fprintf(stderr, "Internal error in create_heyu_environment()\n");
      fprintf(stderr, "Allocated = %d, actual = %d\n",
           newsize, (int)(ep - envp));
      exit(1);
   }

   return envp;
}


/*---------------------------------------------------------------------+
 | Create an environment compatible with Xtend scripts.                |
 +---------------------------------------------------------------------*/
char **create_xtend_environment( int index )
{
   
   extern char   **environ;

   int           j, size, newsize;
   unsigned char hcode, ucode;
   char          **envp, **ep;
   char          hc;
   unsigned int  value, bitmap, addressed, onstate, appliance;
   char          minibuf[512];
   static int    sizchrptr = sizeof(char *);

   static struct {
      char *query;
      int  mask;
   } maskval[] = {
      {"isAppl",    XTMAP_APPL },
      {"isAddr",    XTMAP_ADDR },
      {"isOn",      XTMAP_ON   },
      { NULL,       0          },
   };

   /* Get length of original environment */
   size = 0;
   while ( environ[size] != NULL )
      size++;

   newsize = size + 256 + 50;

   if ( (envp = calloc(newsize, sizchrptr)) == NULL ) {
      fprintf(stderr, "create_xtend_environment() : out_of_memory\n");
      exit(1);
   }

   ep = envp;

   *ep++ = add_envptr("HEYU_MONITOR_CHILD=1");

   *ep++ = add_envptr("IAMXTEND=1");

   /* Put config file pathname in environment for child processes */
   sprintf(minibuf, "X10CONFIG=%s", pathspec(heyu_config));
   *ep++ = add_envptr(minibuf);

   /* Create the various masks */
   j = 0;
   while ( maskval[j].query ) {
      sprintf(minibuf, "%s=%d", maskval[j].query, maskval[j].mask);
      *ep++ = add_envptr(minibuf);
      j++;
   }

   for ( hcode = 0; hcode < 16; hcode++ ) {
      hc = code2hc(hcode);
      addressed = x10state[hcode].addressed;
      onstate = x10state[hcode].state[OnState];
      appliance = ~(modmask[DimMask][hcode] | modmask[BriMask][hcode]);
      for ( ucode = 0; ucode < 16; ucode++ ) {
         bitmap = 1 << ucode;
         value = 0;

         /* Add the state bits */
         if ( appliance & bitmap )
            value |= XTMAP_APPL;         
         if ( addressed & bitmap )
            value |= XTMAP_ADDR;
         if ( onstate & bitmap )
            value |= XTMAP_ON;
            
         sprintf(minibuf, "X10_%c%d=%d", hc, code2unit(ucode), value);
         *ep++ = add_envptr(minibuf);
      }
   }

   /* Append the user's original environment */
   for ( j = 0; j < size; j++ ) {
      *ep++ = add_envptr(environ[j]);
   }

   /* Add NULL terminator */
   *ep++ = NULL;

   return envp;
}

void free_environment ( char **envp )
{
   int j = 0;

   if ( !envp )
      return;
   
   while ( envp[j] != NULL ) {
      free(envp[j]);
      j++;
   }

   free(envp);
   return;
}

/*---------------------------------------------------------------------+
 | For debugging.                                                      |
 +---------------------------------------------------------------------*/
void display_environment ( char **envp ) 
{
   int j = 0;
   
   while ( envp[j] != NULL ) {
      printf("%s\n", envp[j]);
      j++;
   }
}

#ifdef EXEC_POSIX
/*---------------------------------------------------------------------+
 | Launch a general script.                                            |
 +---------------------------------------------------------------------*/
int launch_script ( int *launchp ) 
{

   int      retcode, scriptnum, lindex, pid, status;
   unsigned char option;
   char     *cmdline, *label;
   char     **envp;
   char     *argv[4];
   char     *shell;
   LAUNCHER *launcherp;

   write_x10state_file();

   launcherp = config.launcherp;
   lindex = *launchp;
   scriptnum = launcherp[lindex].scriptnum;
   cmdline = config.scriptp[scriptnum].cmdline;
   option = config.scriptp[scriptnum].script_option;
   label = config.scriptp[scriptnum].label;
   shell = config.script_shell;

   /* We don't want to do this more than once. */
   *launchp = -1;

   fprintf(stdout, "%s Launching '%s': %s\n", statstr, label, cmdline);
   fflush(stdout);

   argv[0] = shell;
   argv[1] = "-c";
   argv[2] = cmdline;
   argv[3] = NULL;

   retcode = fork();

   if ( retcode == -1 ) {
      fprintf(stderr, "Unable to fork() for launching script.\n");
      fflush(stderr);
      return 1;
   }

   if ( retcode == 0 ) {
      /* In child process */
      if ( (pid = fork()) > 0 ) {
	 /* Child dies; grandchild inherited by init */
         _exit(0);
      }
      else if ( pid < 0 ) {
         fprintf(stderr, "Failed to double fork() for launching script.\n");
	 fflush(stderr);
	 _exit(1);
      }
      else {
         if ( option & SCRIPT_NOENV ) 
            envp = create_noenv_environment();
         else if ( option & SCRIPT_XTEND )
            envp = create_xtend_environment(lindex);
         else
            envp = create_heyu_environment(lindex, option);
         execve(shell, argv, envp);
         perror("Execution of script has failed");
         exit(1);
      }
   }

   /* Wait for child process */
   while ( waitpid(retcode, &status, 0) < 0 && errno == EINTR )
      ;
   
   return 0;
}    

/*---------------------------------------------------------------------+
 | Launch heyuhelper.                                                  |
 +---------------------------------------------------------------------*/
int launch_heyuhelper ( unsigned char hcode, unsigned int bitmap,
                                                unsigned char actfunc ) 
{
   int           retcode, pid, status;
   char          *argv[4];
   char          *shell;
   char          **envp;
   char          hc;
   char          cmdline[128];

   write_x10state_file();
   
   hc = tolower(code2hc(hcode));

   sprintf(cmdline, "heyuhelper %c%d%s",
	hc, x10state[hcode].lastunit, funclabel[actfunc]);

   shell = config.script_shell;
   
   argv[0] = shell;
   argv[1] = "-c";
   argv[2] = cmdline;
   argv[3] = 0;

   retcode = fork();

   if ( retcode == -1 ) {
      fprintf(stderr, "Unable to fork() for launching heyuhelper.\n");
      return 1;
   }

   if ( retcode == 0 ) {
      /* In child process */
      if ( (pid = fork()) > 0 ) {
	 /* Child dies; grandchild inherited by init */
         _exit(0); 
      }
      else if ( pid < 0 ) {
	 fprintf(stderr, "Unable to fork() 2nd gen for launching heyuhelper.\n");
	 fflush(stderr);
	 _exit(1);
      }
      else {
	 envp = create_noenv_environment();
         execve(shell, argv, envp);
         /* Silently exit if failure */
         exit(1);
      }
   }

   /* wait for child process */
   while ( waitpid(retcode, &status, 0) < 0 && errno == EINTR )
      ;

   return 0;
}      
      
#else
/*---------------------------------------------------------------------+
 | Launch a general script - non-POSIX                                 |
 +---------------------------------------------------------------------*/
int launch_script ( int *launchp ) 
{

   int      retcode, scriptnum, lindex;
   unsigned char option;
   char     *cmdline, *label;
   char     **envp;
   char     *argv[4];
   char     *shell;
   LAUNCHER *launcherp;

   write_x10state_file();
   
   launcherp = config.launcherp;
   lindex = *launchp;
   scriptnum = launcherp[lindex].scriptnum;
   cmdline = config.scriptp[scriptnum].cmdline;
   option = config.scriptp[scriptnum].script_option;
   label = config.scriptp[scriptnum].label;
   shell = config.script_shell;

   /* We don't want to do this more than once. */
   *launchp = -1;

   fprintf(fdsout, "%s Launching '%s': %s\n", statstr, label, cmdline);
   fflush(fdsout);

   argv[0] = shell;
   argv[1] = "-c";
   argv[2] = cmdline;
   argv[3] = 0;

   retcode = fork();

   if ( retcode == -1 ) {
      fprintf(stderr, "Unable to fork() for launching script.\n");
      return 1;
   }

   if ( retcode == 0 ) {
      /* In child process */
      if ( option & SCRIPT_NOENV )
         envp = create_noenv_environment(); 
      else if ( option & SCRIPT_XTEND )
         envp = create_xtend_environment(lindex);
      else
         envp = create_heyu_environment(lindex, option);
      execve(shell, argv, envp);
      perror("Execution of script has failed");
      exit(1);
   }

   return 0;
}    


/*---------------------------------------------------------------------+
 | Launch heyuhelper - non-POSIX.                                      |
 +---------------------------------------------------------------------*/
int launch_heyuhelper ( unsigned char hcode, unsigned int bitmap,
                                                unsigned char actfunc ) 
{

   int           retcode;
   char          *argv[4];
   char          *shell;
   char          **envp;
   char          hc;
   char          cmdline[128];

   write_x10state_file();
   
   hc = tolower(code2hc(hcode));

   sprintf(cmdline, "heyuhelper %c%d%s",
	hc, x10state[hcode].lastunit, funclabel[actfunc]);

   shell = config.script_shell;
   
   argv[0] = shell;
   argv[1] = "-c";
   argv[2] = cmdline;
   argv[3] = NULL;

   retcode = fork();

   if ( retcode == -1 ) {
      fprintf(stderr, "Unable to fork() for launching heyuhelper.\n");
      return 1;
   }

   if ( retcode == 0 ) {
      /* In child process */
      envp = create_noenv_environment();
      execve(shell, argv, envp);
      /* Silently exit if failure */
      exit(1);
   }

   return 0;
}      
      
#endif  /* End of ifdef EXEC_POSIX / else block */
      

/*---------------------------------------------------------------------+
 | Display state information in a variety of formats (primarily for    |
 | use in scripts) in response to command line requests.               |
 +---------------------------------------------------------------------*/
int c_x10state ( int argc, char *argv[] )
{
   extern void show_module_mask ( unsigned char );

   char          hc;
   unsigned char hcode, ucode, unit;
   unsigned char level;
   unsigned int  bitmap, aflags;

   argv++;
   argc--;

   if ( argc < 1 ) {
      fprintf(stderr, "Too few arguments\n");
      return 1;
   }

   get_configuration();


   if ( strcmp(argv[0], "initstate") == 0 ) {
      /* Initialize the state structure to 0 */
      if ( argc == 1 ) {
         send_x10state_command( ST_INIT_ALL, 0 );
         return 0;
      }
      else {
         aflags = parse_addr(argv[1], &hc, &bitmap);
         if ( !(aflags & A_VALID) || !(aflags & A_HCODE) || (aflags & A_DUMMY) ) {
            fprintf(stderr, "Invalid argument %s\n", argv[1]);
            return 1;
         }
         if ( bitmap ) 
            fprintf(stderr, "Unit code ignored\n");

         send_x10state_command( ST_INIT, hc2code(hc) );
      }
      return 0;
  }

  if ( strcmp(argv[0], "fetchstate") == 0 ) {
      return fetch_x10state();
   }

   if ( strcmp(argv[0], "onstate")     == 0 ||
        strcmp(argv[0], "dimstate")    == 0 ||
        strcmp(argv[0], "addrstate")   == 0 ||
        strcmp(argv[0], "chgstate")    == 0 ||
        strcmp(argv[0], "dimlevel")    == 0 ||
        strcmp(argv[0], "rawlevel")    == 0 ||
        strcmp(argv[0], "heyu_state")  == 0 ||
        strcmp(argv[0], "heyu_rawstate")  == 0 ||
        strcmp(argv[0], "xtend_state") == 0   ) {

      if ( fetch_x10state() != 0 ) {
         return 1;
      }

      if ( argc > 1 ) {
         aflags = parse_addr(argv[1], &hc, &bitmap) ;
         if ( !(aflags & A_VALID) || !(aflags & A_HCODE) || (aflags & A_DUMMY) ) {
            fprintf(stderr, "%s: Invalid Housecode|Unit '%s'\n", argv[0], argv[1]);
            return 1;
         }
         if ( aflags & A_MULT ) {
            fprintf(stderr, "%s: Only a single unit code is acceptable\n", argv[0]);
            return 1;
         }
         if ( !bitmap ) {
            fprintf(stderr, "%s: No unit specified in '%s'\n", argv[0], argv[1]);
            return 1;
         }

         hcode = hc2code(hc);

         hc = tolower(hc);
         ucode = single_bmap_unit(bitmap);
         unit = code2unit(ucode);
         if ( strcmp(argv[0], "onstate") == 0 ) {
            /* heyuhelper format */
            if ( x10state[hcode].state[OnState] & bitmap )
               printf("%c%dOn\n", hc, unit);
            else
               printf("%c%dOff\n", hc, unit);
         }
         else if ( strcmp(argv[0], "dimstate") == 0 ) {
            /* heyuhelper format */
            if ( x10state[hcode].state[DimState] & bitmap )
               printf("%c%dDim\n", hc, unit);
            else if ( x10state[hcode].state[OnState] & bitmap )
               printf("%c%dOn\n", hc, unit);
            else
               printf("%c%dOff\n", hc, unit);
         }
         else if ( strcmp(argv[0], "chgstate") == 0 ) {
            if ( x10state[hcode].state[ChgState] & bitmap )
               printf("%c%dChg\n", hc, unit);
            else
               printf("%c%dUnchg\n", hc, unit);
         }
         else if ( strcmp(argv[0], "addrstate") == 0 ) {
            if ( x10state[hcode].addressed & bitmap )
               printf("%c%dAddr\n", hc, unit);
            else
               printf("%c%dUnaddr\n", hc, unit);
         }
         else if ( strcmp(argv[0], "dimlevel") == 0 ) {
            /* Level as a percentage of full On */
            level = x10state[hcode].dimlevel[ucode];
            if ( modmask[Ext3Mask][hcode] & bitmap )
               printf("%d\n", ext3level2pct(level));
            else if ( modmask[PresetMask][hcode] & bitmap )
               printf("%d\n", presetlevel2pct(level));
            else {
               printf("%d\n", dims2pct(level));
            }
         }
         else if ( strcmp(argv[0], "rawlevel") == 0 ) {
            /* Native module level */
            level = x10state[hcode].dimlevel[ucode];
            if ( modmask[PresetMask][hcode] & bitmap )
               printf("%d\n", level + 1);
            else {
               printf("%d\n", level);
            }
         }
         else if ( strcmp(argv[0], "heyu_state") == 0 ) {
            /* Heyu script bitmap format */
            printf("%d\n", get_heyu_state(hcode, ucode));
         } 
         else if ( strcmp(argv[0], "heyu_rawstate") == 0 ) {
            /* Heyu script bitmap format, but with native levels */
            printf("%d\n", get_heyu_rawstate(hcode, ucode));
         } 
         else if ( strcmp(argv[0], "xtend_state") == 0 ) {
            /* Xtend script bitmap format */
            printf("%d\n", get_xtend_state(hcode, ucode));
         } 

         return 0;
      }
      fprintf(stderr, "%s: Housecode|Unit needed\n", argv[0]);
      return 1;
   }

   if ( strcmp(argv[0], "statestr") == 0 ) {

      if ( fetch_x10state() != 0 ) {
         return 1;
      }
      if ( argc > 1 ) {
         aflags = parse_addr(argv[1], &hc, &bitmap) ;
         if ( !(aflags & A_VALID) || !(aflags & A_HCODE) || (aflags & A_DUMMY) ) {
            fprintf(stderr, "%s: Invalid Housecode in '%s'\n", argv[0], argv[1]);
            return 1;
         }
         if ( bitmap ) {
            fprintf(stderr, "%s: Units in '%s' ignored\n", argv[0], argv[1]);
         }

         hcode = hc2code(hc);
         hc = toupper(hc);

         printf("%c%s\n", hc,
           bmap2asc2(x10state[hcode].state[OnState], x10state[hcode].state[DimState], "013"));

         return 0;
      }
      fprintf(stderr, "%s: Housecode needed\n", argv[0]);
      return 1;
   }


   fprintf(stderr, "State command '%s' not recognized\n", argv[0]);
   return 1;
}

/*---------------------------------------------------------------------+
 | Display a table showing the dimlevels of every Housecode|Unit,      |
 | expressed as a (integer) percentage of full On brightness.          |
 +---------------------------------------------------------------------*/
void show_all_dimlevels ( void )
{
   char hc;
   int  unit;
   unsigned char hcode, ucode, level;
   unsigned int mask;

   printf("%20sPercent of Full Brightness\n", "");
   printf("  Unit:");
   for ( unit = 1; unit <= 16; unit++ )
      printf("%3d ", unit);
   printf("\n");
   printf("Hcode  ");
   for ( unit = 1; unit <= 16; unit++ )
      printf(" -- ");
   printf("\n");
   
   for ( hc = 'A'; hc <= 'P'; hc++ ) {
      printf("  %c    ", hc);
      hcode = hc2code(hc);
      for ( unit = 1; unit <= 16; unit++ ) {
         ucode = unit2code(unit);
         mask = 1 << ucode;
         level = x10state[hcode].dimlevel[ucode];
         if ( modmask[Ext3Mask][hcode] & mask )
            printf("%3d ", ext3level2pct(level));
         else if ( modmask[PresetMask][hcode] & mask )
            printf("%3d ", presetlevel2pct(level));
         else {
            printf("%3d ", dims2pct(level));
         }
      }
      printf("\n");
   }

   return;
}

/*---------------------------------------------------------------------+
 | Display a table showing the dimlevel of each Housecode|Unit in the  |
 | native form for the module at the address, i.e., 0-210 for standard |
 | modules, 1-32 for Preset modules, and 0-63 for Extended modules.    |
 | Preset and Extended levels are prefixed respectively by 'p' and 'e'.|
 +---------------------------------------------------------------------*/
void show_all_dimlevels_raw ( void )
{
   char hc;
   int  unit;
   unsigned char hcode, ucode, level;
   unsigned int mask;
   char buffer[16];

   printf("Raw Brightness Levels (p: Preset 1-32, e: Extended 0-63, else 0-210)\n");
   printf("  Unit:");
   for ( unit = 1; unit <= 16; unit++ )
      printf("%3d ", unit);
   printf("\n");
   printf("Hcode  ");
   for ( unit = 1; unit <= 16; unit++ )
      printf(" -- ");
   printf("\n");
   
   for ( hc = 'A'; hc <= 'P'; hc++ ) {
      printf("  %c    ", hc);
      hcode = hc2code(hc);
      for ( unit = 1; unit <= 16; unit++ ) {
         ucode = unit2code(unit);
         mask = 1 << ucode;
         level = x10state[hcode].dimlevel[ucode];
         if ( modmask[Ext3Mask][hcode] & mask ) 
            sprintf(buffer, "e%d", level);
         else if ( modmask[PresetMask][hcode] & mask )
            sprintf(buffer, "p%d", level + 1);
         else 
            sprintf(buffer, "%d", level);
         printf("%3s ", buffer);
      }
      printf("\n");
   }

   return;
}


/*---------------------------------------------------------------------+
 | Return the index to a new LAUNCHER in the array of same.            |
 +---------------------------------------------------------------------*/
int launcher_index ( LAUNCHER **launcherpp ) 
{
   static int    size, maxsize;
   static int    sizlauncher = sizeof(LAUNCHER);
   int           j, index, blksize = 5;

   /* Allocate initial block if not already done */
   if ( *launcherpp == NULL ) {
      *launcherpp = calloc(blksize, sizlauncher );
      if ( *launcherpp == NULL ) {
         (void) fprintf(stderr, "Unable to allocate memory for Launcher.\n");
         exit(1);
      }
      maxsize = blksize;
      size = 0;
      for ( j = 0; j < maxsize; j++ ) {
         (*launcherpp)[j].line_no = -1 ;
         (*launcherpp)[j].label[0] = '\0';
         (*launcherpp)[j].bmaptrig = 0;
         (*launcherpp)[j].chgtrig = 0;
         (*launcherpp)[j].afuncmap = 0;
         (*launcherpp)[j].gfuncmap = 0;
         (*launcherpp)[j].signal = 0;
         (*launcherpp)[j].module = 0;
         (*launcherpp)[j].source = 0;
         (*launcherpp)[j].oksofar = 0;
      }
   }

   /* Check to see if there's an available location          */
   /* If not, increase the size of the memory allocation.    */
   /* (Always leave room for a final termination indicator.) */
   if ( size == (maxsize - 1) ) {
      maxsize += blksize ;
      *launcherpp = realloc(*launcherpp, maxsize * sizlauncher );
      if ( *launcherpp == NULL ) {
         (void) fprintf(stderr, "Unable to increase size of Launcher list.\n");
         exit(1);
      }

      /* Initialize the new memory allocation */
      for ( j = size; j < maxsize; j++ ) {
         (*launcherpp)[j].line_no = -1 ;
         (*launcherpp)[j].label[0] = '\0';
         (*launcherpp)[j].bmaptrig = 0;
         (*launcherpp)[j].chgtrig = 0;
         (*launcherpp)[j].afuncmap = 0;
         (*launcherpp)[j].gfuncmap = 0;
         (*launcherpp)[j].signal = 0;
         (*launcherpp)[j].module = 0;
         (*launcherpp)[j].source = 0;
         (*launcherpp)[j].oksofar = 0;
      }
   }

   index = size;
   size += 1;
   return index;
}

/*---------------------------------------------------------------------+
 | Create individual launchers with argument label from the sets of    |
 | launch conditions in argument tail.  Return 0 if successful, or -1  |
 | if error.                                                           |
 +---------------------------------------------------------------------*/
int create_launchers ( LAUNCHER **launcherpp, char *label, int line_no, char *tail ) 
{

   int           j, k, m, index;  
   int           argc, tokc;
   char          **argv, **tokv;
   char          *sp;
   char          hc;
   char          errmsg[128];
   unsigned char aflags, hcode, change;
   unsigned char actual, generic, signal, module, source, nosource;
   unsigned int  bitmap;
   char          tbuf[32];

   sp = tail;

   /* Tokenize the line into individual sets of launch conditions */
   tokenize( sp, ";", &argc, &argv );

   if ( argc == 0 ) {
       store_error_message("No launch parameters specified");
       free(argv);
       return -1;
   }

   for ( j = 0; j < argc; j++ ) {
      /* Tokenize each set of launch conditions */
      tokenize(argv[j], " \t", &tokc, &tokv);
      if ( tokc < 2 ) {
         sprintf(errmsg, "No launch parameters specified for %s", tokv[0]);
         store_error_message(errmsg);
         free(tokv);
         free(argv);
         return -1;
      }

      index = launcher_index(launcherpp);

      (*launcherpp)[index].line_no = line_no;
      strcpy((*launcherpp)[index].label, label);

      change = 0; module = 0; signal = 0; source = 0; nosource = 0;
      actual = 0; generic = 0;

      aflags = parse_addr(tokv[0], &hc, &bitmap);
      if ( !(aflags & A_VALID) || !(aflags & A_HCODE) || !bitmap ) {
         sprintf(errmsg, "Invalid Housecode|Units '%s' in Launcher.", tokv[0]);
         store_error_message(errmsg);
         free(tokv);
         free(argv);
         return -1;
      }
      hcode = hc2code(hc);
      (*launcherpp)[index].hcode = hc2code(hc);
      (*launcherpp)[index].bmaptrig = bitmap;

      /* Scan the remainder of the tokens for keywords */
      for ( k = 1; k < tokc; k++ ) {
         strncpy2(tbuf, tokv[k], (sizeof(tbuf)));
         strlower(tbuf);

         if ( strcmp("changed", tbuf) == 0 ) {
            change = 1;
            *tokv[k] = '\0';
         }
         else if ( strcmp("module", tbuf) == 0 ) {
            module = 1;
            *tokv[k] = '\0';
         }
         else if ( strcmp("nosrc", tbuf) == 0 ) {
            source |= LSNONE; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("anysrc", tbuf) == 0 ) {
            source |= LSALL; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("sndc", tbuf) == 0 ) {
            source |= SNDC; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("snds", tbuf) == 0 ) {
            source |= SNDS; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("sndm", tbuf) == 0 ) {
            source |= SNDM; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("sndt", tbuf) == 0 ) {
            source |= SNDT; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("rcvi", tbuf) == 0 ) {
            source |= RCVI; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("rcvt", tbuf) == 0 ) {
            source |= RCVT; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("nosndc", tbuf) == 0 ) {
            nosource |= SNDC; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("nosnds", tbuf) == 0 ) {
            nosource |= SNDC; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("nosndm", tbuf) == 0 ) {
            nosource |= SNDM; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("nosndt", tbuf) == 0 ) {
            nosource |= SNDT; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("norcvi", tbuf) == 0 ) {
            nosource |= RCVI; 
            *tokv[k] = '\0';
         }
         else if ( strcmp("norcvt", tbuf) == 0 ) {
            nosource |= RCVT; 
            *tokv[k] = '\0';
         }
      }
         
      for ( k = 1; k < tokc; k++ ) {
         if ( *tokv[k] == '\0' )
            continue;
         for ( m = 0; m < NTRIGTYPES; m++ ) {
            if ( strcmp(trig_type[m].label, tokv[k]) == 0 ) {
               if ( trig_type[m].generic ) {
                  (*launcherpp)[index].gfuncmap |= (1 << trig_type[m].signal);
               }
               else {
                  (*launcherpp)[index].afuncmap |= (1 << trig_type[m].signal);
               }

               break;
            }
         }
         if ( m >= NTRIGTYPES ) {
            sprintf(errmsg, "Invalid launch parameter '%s'", tokv[k]);
            store_error_message(errmsg);
            free(tokv);
            free(argv);
            return -1;
         }
      }


      if ( change ) {
         (*launcherpp)[index].chgtrig = (*launcherpp)[index].bmaptrig;
         module = 1;
         signal = 0;
      }
      else
         (*launcherpp)[index].chgtrig = 0;

      if ( module ) {
         (*launcherpp)[index].module = bitmap;
         (*launcherpp)[index].signal = 0;
      }
      else if ( signal ) {
         (*launcherpp)[index].module = 0;
         (*launcherpp)[index].signal = bitmap;
      }
      else {
         (*launcherpp)[index].module = 0;
         (*launcherpp)[index].signal = 0;
      }

      (*launcherpp)[index].source = source;
      (*launcherpp)[index].nosource = nosource;

      (*launcherpp)[index].oksofar = 1;

      free(tokv);
   }
   free(argv);

   return 0;
} 


/*---------------------------------------------------------------------+
 | Add a launcher to the array of LAUNCHER structures for each set     |
 | of launch conditions in argument tail.  Return 0 if successful.     | 
 +---------------------------------------------------------------------*/
int add_launchers ( LAUNCHER **launcherpp, int line_no, char *tail ) 
{

   char          label[51];
   char          errmsg[128];
   char          *sp;

   sp = tail;

   /* The first token is the script label */
   if ( *get_token(label, &sp, " \t", 50) == '\0' || *sp == '\0' ) {
      store_error_message("Too few items on line.");
      return -1;
   }
   if ( strlen(label) > NAME_LEN ) {
      sprintf(errmsg, "Label '%s' too long (max %d characters)", label, NAME_LEN);
      store_error_message(errmsg);
      return -1;
   }

   strtrim(sp);
   if ( *sp == '\0' ) {
      store_error_message("Too few items on line.");
      return -1;
   }

   return create_launchers(launcherpp, label, line_no, sp);
}


/*---------------------------------------------------------------------+
 | Make the appropriate changes/additions/error checks for launchers   |
 | once all the config file information has been stored.               |
 +---------------------------------------------------------------------*/
int finalize_launchers ( void )
{
   unsigned char mode, matched;
   int           j, k, errors = 0;
   LAUNCHER      *launcherp;
   SCRIPT        *scriptp;

   mode = config.launch_mode ;
   launcherp = config.launcherp;
   scriptp = config.scriptp;

   if ( !launcherp && !scriptp )
      return 0;
   else if ( launcherp && !scriptp ) {
      store_error_message("No SCRIPTs defined for LAUNCHERs");
      return 1;
   }
   else if ( !launcherp && scriptp ) {
      store_error_message("No LAUNCHERs defined for SCRIPTs");
      return 1;
   } 

   errors = 0;
   j = 0;
   while ( launcherp[j].line_no > 0 ) {
      /* Skip known bad launchers */
      if ( !(launcherp[j].oksofar) ) {
         j++;
         continue;
      }

      /* Consolidate signal sources */
      if ( launcherp[j].source & LSNONE ) 
         launcherp[j].source &= ~(launcherp[j].nosource | LSNONE);
      else
         launcherp[j].source =
           (config.launch_source | launcherp[j].source) & ~launcherp[j].nosource;

      if ( !launcherp[j].source ) {
         fprintf(stderr,
            "Config Line %02d: No sources (e.g., RCVI) for LAUNCHER with label '%s'\n",
                    launcherp[j].line_no, launcherp[j].label);
         errors++;
      }
         

      if ( launcherp[j].signal )
         launcherp[j].signal = launcherp[j].bmaptrig;
      else if ( launcherp[j].module )
         launcherp[j].signal = 0 /* ~launcherp[j].bmaptrig */;
      else if ( mode == TMODE_SIGNAL ) 
         launcherp[j].signal = launcherp[j].bmaptrig;
      else
         launcherp[j].signal = 0 /* ~launcherp[j].bmaptrig */;


      /* Fix up 'changed' function if no other function specified */
      if ( launcherp[j].chgtrig != 0 ) {
         if ( launcherp[j].afuncmap == 0 && launcherp[j].gfuncmap == 0 )
            launcherp[j].afuncmap = ~(0);
      }  

      k = 0;
      matched = 0;
      while ( scriptp[k].line_no > 0 ) {
         if ( strcmp(launcherp[j].label, scriptp[k].label) == 0 ) {
            launcherp[j].scriptnum = k;
            matched = 1;
            break;
         }
         k++;
      }
      if ( !matched ) {
         fprintf(stderr, "Config Line %02d: No SCRIPT for LAUNCHER with label '%s'\n",
                    launcherp[j].line_no, launcherp[j].label);
         errors++;
      }
      j++;
   }

   k = 0;
   while ( scriptp[k].line_no > 0 ) {
      j = 0;
      matched = 0;
      while ( launcherp[j].line_no > 0 ) {
         if ( strcmp(launcherp[j].label, scriptp[k].label) == 0 ) {
            matched = 1;
            break;
         }
         j++;
      }
      if ( !matched ) {
         fprintf(stderr, "Config Line %02d: No LAUNCHER for SCRIPT with label '%s'\n", 
                 scriptp[k].line_no, scriptp[k].label);
         errors++;
      }
      k++;
   }

   return errors;
}

/*---------------------------------------------------------------------+
 | Add a script to the array of SCRIPT structures and return           |
 | the index of the new member.                                        |
 +---------------------------------------------------------------------*/
int add_script ( SCRIPT **scriptpp, LAUNCHER **launcherpp, int line_no, char *tail )
{
   static int size, maxsize;
   static int sizscript = sizeof(SCRIPT);
   int        blksize = 5;
   int        j, index;
   char       *sp, *dup;
   char       token[51];
   char       errmsg[128];
   char       *separator = "::";

   /* Allocate memory for script structure */

   /* Allocate initial block if not already done */
   if ( *scriptpp == NULL ) {
      *scriptpp = calloc(blksize, sizscript );
      if ( *scriptpp == NULL ) {
         (void) fprintf(stderr, "Unable to allocate memory for Script.\n");
         exit(1);
      }
      maxsize = blksize;
      size = 0;
      for ( j = 0; j < maxsize; j++ ) {
         (*scriptpp)[j].line_no = -1 ;
         (*scriptpp)[j].label[0] = '\0';
         (*scriptpp)[j].script_option = 0;
         (*scriptpp)[j].cmdline = NULL;
      }
   }

   /* Check to see if there's an available location          */
   /* If not, increase the size of the memory allocation.    */
   /* (Always leave room for a final termination indicator.) */
   if ( size == (maxsize - 1) ) {
      maxsize += blksize ;
      *scriptpp = realloc(*scriptpp, maxsize * sizscript );
      if ( *scriptpp == NULL ) {
         (void) fprintf(stderr, "Unable to increase size of Script list.\n");
         exit(1);
      }

      /* Initialize the new memory allocation */
      for ( j = size; j < maxsize; j++ ) {
         (*scriptpp)[j].line_no = -1 ;
         (*scriptpp)[j].label[0] = '\0';
         (*scriptpp)[j].script_option = 0;
         (*scriptpp)[j].cmdline = NULL;
      }
   }

   index = size;
   size += 1;

   /* Now add the new script information */

   sp = tail;

   (*scriptpp)[index].line_no = line_no;

   /* Locate substring separating the launcher from the command line */
   if ( (sp = strstr(tail, separator)) == NULL ) {
      sprintf(errmsg, "Invalid SCRIPT format - no '%s' separator\n", separator);
      store_error_message(errmsg);
      return -1;
   }

   *sp = '\0';
    sp += strlen(separator);

   strtrim(sp);
   if ( *sp == '\0' ) {
      store_error_message("Missing script command line\n");
      return -1;
   }
   
   /* Set the option for alias states in the heyu environment */
   (*scriptpp)[index].script_option |= ENV_ALIAS;

   /* Look for -xtend option */
   if ( strncmp(sp, "-x", 2) == 0 ) {
      (*scriptpp)[index].script_option |= SCRIPT_XTEND;
      (*scriptpp)[index].script_option &= ~ENV_ALIAS;
      /* Bypass the token */
      get_token(token, &sp, " \t", 50);
      strtrim(sp);
   }

   /* Look for -noenv option */
   if ( strncmp(sp, "-n", 2) == 0 ) {
      (*scriptpp)[index].script_option |= SCRIPT_NOENV;
      /* Bypass the token */
      get_token(token, &sp, " \t", 50);
      strtrim(sp);
   }

   /* Look for -rawlevel option */
   if ( strncmp(sp, "-r", 2) == 0 ) {
      (*scriptpp)[index].script_option |= LEVEL_RAW;
      /* Bypass the token */
      get_token(token, &sp, " \t", 50);
      strtrim(sp);
   }

   if ( *sp == '\0' ) {
      store_error_message("Missing script command line\n");
      return -1;
   }

   if ( (dup = strdup(sp)) == NULL ) {
      fprintf(stderr, "Unable to allocate memory for script command line.\n");
      exit(1);
   }
   (*scriptpp)[index].cmdline = dup;

   sp = tail;
   
   /* Does script have a label? */
   get_token(token, &sp, " \t", 50);
   if ( strcmp(token, "-l") == 0 ) {
      /* Yes */
      get_token(token, &sp, " \t", 50);
      strcpy((*scriptpp)[index].label, token);
   }
   else {
      /* No - create default label */
      sprintf((*scriptpp)[index].label, "Script_%02d", line_no);
      sp = tail;
   }

   strtrim(sp);

   /* Add launchers if specified here, otherwise assume user will */
   /* add them as a separate config item.                         */
   if ( *sp != '\0' ) {
      create_launchers (launcherpp, (*scriptpp)[index].label, line_no, sp);
   }

   return index;
}
      

/*---------------------------------------------------------------------+ 
 | Load the x10image file and compare it's checksum with the argument. |
 | Return 1 if OK, 0 otherwise.                                        |
 +---------------------------------------------------------------------*/
int load_image ( int chksum )
{

   FILE *fd;
   int  sum;

   if ( !(fd = fopen(pathspec(IMAGE_FILE), "r")) ) {
      return 0;
   }

   if ( fread((void *)image, 1, PROMSIZE + 2, fd) != PROMSIZE ) {
      fclose(fd);
      return 0;
   }
   fclose(fd);

   sum = image_chksum(image);
   if ( sum != chksum || sum == -1 )
      return 0;

   return 1;
}

/*---------------------------------------------------------------------+
 | Return a pointer to the (static) image.                             |
 +---------------------------------------------------------------------*/
unsigned char *image_ptr ( void )
{
   return image;
}

/*---------------------------------------------------------------------+ 
 | Scan the cm11a memory image to locate the trigger having argument   |
 | tag which resulted in execution of the macro at macaddr.  Pass back |
 | the housecode|unit and housecode|function bytes through the         |
 | argument list.  Return the tag (1-6) of the matching trigger, or 0  |
 | if no match or if the tag is 7 (indicating anything more than 6).   |
 +---------------------------------------------------------------------*/
int find_trigger ( unsigned char *image, unsigned int macaddr,
    unsigned char tag, unsigned char *hcu, unsigned char *hcf,
    unsigned char *fwver )
{
   unsigned char func, count, *sp;
   unsigned int  addr;

   if ( tag > 6 )
      return 0;

   count = 0;

   /* Start of trigger table */
   addr = ((image[0] << 8) | image[1]) & 0x3ff;
   sp = image + addr;

   /* Future:                                          */
   /* Heyu stores CM11A firmware version in a byte     */
   /* between the timer table terminator 0xff and the  */
   /* start of the trigger table.                      */
   *fwver = *(sp - 1); 

   while ( sp < (image + PROMSIZE - 2) ) {
      if ( *sp == 0xff && *(sp+1) == 0xff )
         break;
      addr = (((*(sp+1)) << 8) | *(sp+2)) & 0x7ff;
      if ( addr == macaddr && ++count == tag ) {
         *hcu = *sp;
         func = (*(sp+1) & 0x80 ) ? 2 : 3;
         *hcf = (*hcu & 0xf0) | func;
         return count;
      }
      sp += 3;
   }

   return 0;
}

/*---------------------------------------------------------------------+
 | Display the trigger with argument tag which executed the macro at   |
 | macaddr.  Also update the system state and launch a script if so    |
 | programmed.                                                         | 
 +---------------------------------------------------------------------*/
void display_trigger( char *statstr, unsigned int macaddr, unsigned char tag )
{
   unsigned char hcu, hcf, fwver, buf[4];
   int           launchp;
   int           count;

   count = find_trigger(image, macaddr, tag, &hcu, &hcf, &fwver);

   /* Just return if macro delay > 0 */
   if ( image[macaddr] > 0 )
      return;

   if ( count > 0 ) {
      buf[0] = 0x04;
      buf[1] = hcu;
      fprintf(fdsout, "%s rcvt %s\n", statstr, translate_sent(buf, 2, &launchp));
      buf[0] = 0x06;
      buf[1] = hcf;
      fprintf(fdsout, "%s rcvt %s\n", statstr, translate_sent(buf, 2, &launchp));
      if ( launchp >= 0 )
         launch_script(&launchp);
   }
   else {
      fprintf(fdsout, "%s rcvt Undetermined trigger\n", statstr);
   }

   return;
}

/*---------------------------------------------------------------------+
 |  Perform multiple functions when a macro is executed:               |
 |     Display individual addresses/functions in monitor               |
 |     Update x10status for each address/function                      |
 |     If trigger conditions are met, launch the user's script.        |
 |                                                                     |
 |  Argument 'type' is TRIGGER_EXEC or TIMER_EXEC                      |
 +---------------------------------------------------------------------*/
int expand_macro( char *statstr, unsigned int macaddr, int type )
{
   int           nelem, j;
   unsigned char hcode, ucode, level, len;
   unsigned int  bitmap, mask;
   unsigned char buf[8];
   unsigned char *elemp;
   int           launchp;
   char          *pfix;

   static unsigned char cmdlen[16] = {
      3,3,3,3,4,4,3,6,3,3,3,3,3,3,3,3
   };

   if ( type == TRIGGER_EXEC ) 
      pfix = "sndt";
   else
      pfix = "sndm"; 

   nelem = image[macaddr + 1];
   elemp = image + macaddr + 2;

   for ( j = 0; j < nelem; j++ ) {
      hcode = (elemp[0] & 0xf0) >> 4;
      len = cmdlen[elemp[0] & 0x0f];

      bitmap = (elemp[1] << 8) | elemp[2];

      /* Display addresses */
      mask = 1;
      for ( ucode = 0; ucode < 16; ucode++ ) {
         if ( bitmap & mask ) {
            buf[0] = 0x04;
            buf[1] = (hcode << 4) | ucode ;
            fprintf(fdsout, "%s %s %s\n", statstr, pfix, translate_sent(buf, 2, &launchp));
         }
         mask = mask << 1;
      }

      if ( len == 3 ) {
         /* Basic command */
         buf[0] = 0x06;
         buf[1] = elemp[0];
         fprintf(fdsout, "%s %s %s\n", statstr, pfix, translate_sent(buf, 2, &launchp));
         if ( launchp >= 0 )
            launch_script(&launchp);
      }
      else if ( len == 4 ) {
         /* Dim or Bright command */
         level = elemp[3] & 0x1f;
         if ( elemp[3] & 0x80 ) {
            /* Brighten bit is set */
            buf[0] = (22 << 3) | 0x06;
            buf[1] = (hcode << 4) | 5;
            fprintf(fdsout, "%s %s %s\n", statstr, pfix, translate_sent(buf, 2, &launchp));
            if ( launchp >= 0 )
               launch_script(&launchp);
         }
         buf[0] = (level << 3) | 0x06;
         buf[1] = elemp[0];
         fprintf(fdsout, "%s %s %s\n", statstr, pfix, translate_sent(buf, 2, &launchp));
         if ( launchp >= 0 )
            launch_script(&launchp);
      }
      else {
         /* Extended command */
         buf[0] = 0x07;
         buf[1] = elemp[0];
         buf[2] = elemp[3];
         buf[3] = elemp[4];
         buf[4] = elemp[5];
         fprintf(fdsout, "%s %s %s\n", statstr, pfix, translate_sent(buf, 5, &launchp));
         if ( launchp >= 0 )
            launch_script(&launchp);
      }
      elemp += len;
   }
   return 1;
}

/*---------------------------------------------------------------------+
 | Write a lock file for the state engine.  Return 1 if one already    |
 | exists or can't be written; 0 otherwise.                            |
 +---------------------------------------------------------------------*/
int lock_state_engine ( pid_t pid )
{
   FILE *fdlock;
   int  acc;
   char buffer[PATH_LEN + 1];

   sprintf(buffer, "%s/LCK..%s%s", LOCKDIR, X10STATE_FILE, config.suffix);
   if ( (acc = access(buffer, F_OK)) == 0 ) {
      /* File exists - assume state engine is running */
      printf("Heyu state engine is already running.\n");
      return 1;
   }

   if ( (fdlock = fopen(buffer, "w")) == NULL ) {
      fprintf(stderr, "Unable to lock heyu state engine.");
      return 1;
   }

   fprintf(fdlock, "%d\n", (int)pid);
   fclose(fdlock);

   return 0;
}

/*---------------------------------------------------------------------+
 | Unlink the lock file for the state engine.                          |
 +---------------------------------------------------------------------*/
int unlock_state_engine ( void )
{
   char buffer[32];

   sprintf(buffer, "%s/LCK..%s%s", LOCKDIR, X10STATE_FILE, config.suffix);

   unlink(buffer);

   return 0;
}

/*---------------------------------------------------------------------+
 | Return 1 if a lock file exists for the state engine, 0 otherwise.   |
 +---------------------------------------------------------------------*/
int checklock_state_engine ( void )
{
   char buffer[32];
   int  acc;

   sprintf(buffer, "%s/LCK..%s%s", LOCKDIR, X10STATE_FILE, config.suffix);
   if ( (acc = access(buffer, F_OK)) == 0 ) {
      return 1;
   }
   return 0;
}

/*---------------------------------------------------------------------+
 | Tell the state engine daemon to shut itself down and wait for it to |
 | unlink the lock file.  This will fail if called by heyu on a        |
 | different channel (serial_port) since the spool file is different.  |
 | Return 0 if successful or non-zero otherwise.                       |
 +---------------------------------------------------------------------*/
int stop_state_engine ( void )
{
   int  j, k, retcode = 1;

   if ( (retcode = checklock_state_engine()) != 0 ) {
      for ( j = 0; j < 3; j++ ) {
         send_x10state_command(ST_EXIT, 0);
         for ( k = 0; k < 10; k++ ) {
            millisleep(100);
            if ( (retcode = checklock_state_engine()) == 0 )
               break;
         }
         if ( retcode == 0 )
            break;
      }
   }

   return retcode;
}


/*---------------------------------------------------------------------+
 | Start or restart the state engine process.                          |
 +---------------------------------------------------------------------*/
int c_start_engine ( int argc, char *argv[] )
{
   extern char   **environ;

   int           retcode;
   char          *aargv[4];
   char          *shell, *dup;
   char          buffer[127];
   char          tmpfile[PATH_LEN + 1];
   pid_t         pid;

   /* Stop an existing state engine */
   if ( (retcode = stop_state_engine()) != 0 ) {
      fprintf(stderr, "Only master heyu channel can restart state engine.\n");
      exit(1);
   }

#if 0
   if ( (retcode = checklock_state_engine()) != 0 ) {
      send_x10state_command(ST_EXIT, 0);
      for ( j = 0; j < 20; j++ ) {
         millisleep(100);
         if ( (retcode = checklock_state_engine()) == 0 )
            break;
      }
      if ( retcode != 0 ) {
         fprintf(stderr, "This heyu channel cannot restart state engine.\n");
         exit(1);
      }
   }

   if ( checklock_state_engine() != 0 ) {
      /* An engine is already running, tell it to */
      /* exit so we can start a new one.          */
      unlock_state_engine();
      send_x10state_command(ST_EXIT, 0);
   }
#endif

   if ( (shell = getenv("SHELL")) == NULL )
      shell = "/bin/sh";      

   sprintf(buffer, "%s monitor -engine", argv[0]);
   
   aargv[0] = shell;
   aargv[1] = "-c";
   aargv[2] = buffer;
   aargv[3] = NULL;

   retcode = fork();

   if ( retcode == -1 ) {
      fprintf(stderr, "Unable to fork() for starting state engine.\n");
      return 1;
   }

   if ( retcode == 0 ) {
      /* Put config path into environment for any heyu child process */
      sprintf(tmpfile, "X10CONFIG=%s", pathspec(heyu_config));
      if ( (dup = strdup(tmpfile)) == NULL || putenv(dup) == -1 ) {
         fprintf(stderr, "Memory error in c_start_engine()\n");
         exit(1);
      }      
      pid = getpid();
      lock_state_engine(pid);
      /* In child process */
      execve(shell, aargv, environ);
      /* Exit if failure */
      fprintf(stderr, "Unable to start state engine\n");
      unlock_state_engine();
      exit(1);
   }

   return 0;
}      
      
   
